unit Dib256;
{*
   This unit provides a way of loading/saving or even creating bitmap
files with 8 bits per pixel. It suports loading of compressed bitmaps
but not saving them.
   Some bitmap file formats are not supported so if you you have any
problems reading a bitmap file just try to load and save it in a drawing
application like Paint Shop Pro or Corel Photo-Paint.
   It also contains some general functions to copy sprites and
compress/decompress bitmaps.
*}

interface

uses
  Windows, SysUtils, Classes, Graphics, WinG, Dialogs;

type
  TRGBQUAD256 = array [0..255] of TRGBQUAD;
  PRGBQUAD256 = ^TRGBQUAD256;
  TDIB256INFO = record
    bmiHeader: TBITMAPINFOHEADER;
    bmiColors: TRGBQUAD256;
  end;
  PDIB256INFO = ^TDIB256INFO;

  TIOEvent = procedure (Sender: TObject; CompletedPercent: Byte) of object;

   TDIB256 = class( TObject )
   private
      ptDib: PDIB256INFO;
      FWidth: Word;
      FHeight: Word;
      FOnLoading: TIOEvent;
      FOnSaving: TIOEvent;

      procedure SetWidth( Value: Word );
      procedure SetHeight( Value: Word );
      function GetColorTable: PRGBQUAD256;
      function GetPixels: Pointer;
      function GetWidthBytes: Word;
      function GetImageSize: Longint;
      procedure SetPixelAt(X, Y: Integer; Value: Byte);
      function GetPixelAt(X, Y: Integer): Byte;
      procedure CreateDIB;
      procedure DeleteDIB;
      procedure ReadBitmapInfo( fs: TStream; pD: PDIB256INFO );
      procedure InvalidBitmap;

   protected

   public
      constructor Create;
      destructor Destroy; override;

      procedure LoadFromFile( const FileName: string );
      procedure LoadFromStream( Stream: TStream );
      procedure SaveToFile( const FileName: string );
      procedure SaveToStream( Stream: TStream );
      procedure FlipImage; // turns the image upside down
      procedure Copy(Dest: TDib256; SrcX, SrcY, DestX, DestY, CopyWidth,
        CopyHeight: Integer);
      procedure TransCopy(Dest: TDib256; SrcX, SrcY, DestX, DestY,
        CopyWidth, CopyHeight: Integer; TransColor: Byte);
      procedure CopyToWinG(Dest: TWinG; SrcX, SrcY, DestX, DestY,
        CopyWidth, CopyHeight: Integer);
      procedure TransCopyToWinG(Dest: TWinG; SrcX, SrcY, DestX, DestY,
        CopyWidth, CopyHeight: Integer; TransColor: Byte);
      procedure CopyFromWinG(Source: TWinG; SrcX, SrcY, DestX, DestY, CopyWidth,
        CopyHeight: Integer);
      procedure TransCopyFromWinG(Source: TWinG; SrcX, SrcY, DestX, DestY,
        CopyWidth, CopyHeight: Integer; TransColor: Byte);

      property Width: Word read FWidth write SetWidth;
      property Height: Word read FHeight write SetHeight;
      property Dib: PDIB256INFO read ptDib;
      property ColorTable: PRGBQUAD256 read GetColorTable;
      property PixelAt[X,Y: Integer]: Byte read GetPixelAt write SetPixelAt;
      property Pixels: Pointer read GetPixels;
      property WidthBytes: Word read GetWidthBytes;   // width 4 byte aligned
      property ImageSize: Longint read GetImageSize;

      property OnLoading: TIOEvent read FOnLoading write FOnLoading;
      property OnSaving: TIOEvent read FOnSaving write FOnSaving;
   end;

   procedure DecompressBitmap( pDest, pSource: Pointer; wWidth: Word );
   function CompressBitmap( pSource, pDest: Pointer; NBytes: Longint ): Longint;
   procedure TransCopyDIBBits( pDest, pSource: Pointer;
      dwWidth, dwHeight: Longint; dwScanD, dwScanS: Longint; bTranClr: Byte );
   procedure CopyDIBBits( pDest, pSource: Pointer; dwWidth, dwHeight: Longint;
      dwScanD, dwScanS: Longint );

const
  BI_BITFIELDS = 3;

  IOBLOCK_SIZE = 65536;

	{* flags for _lseek *}
	SEEK_SET = 0;
	SEEK_CUR = 1;
	SEEK_END = 2;


implementation

constructor TDIB256.Create;
begin
   inherited Create;

{REG}
end;


destructor TDIB256.Destroy;
begin
   DeleteDib;

   inherited Destroy;
end;


{***** Properties *****}

procedure TDIB256.SetWidth( Value: Word );
begin
   if (FWidth = Value) or (Value < 1) then
      exit;

   FWidth := Value;
   CreateDib;
end;


procedure TDIB256.SetHeight( Value: Word );
begin
   if (FHeight = Value) or (Value < 1) then
      exit;

   FHeight := Value;
   CreateDib;
end;


function TDIB256.GetImageSize: LongInt;
begin
   if (ptDib = nil) then
      Result := 0
   else
      Result := ptDib^.bmiHeader.biSizeImage;
end;


function TDIB256.GetWidthBytes: Word;
begin
   if (ptDib = nil) then
      Result := 0
   else
      Result := (ptDib^.bmiHeader.biWidth+3) and $FFFC;
end;



function TDIB256.GetPixels: Pointer;
begin
   if (ptDib = nil) then
      Result := nil
   else
      Result := Pointer(PChar(ptDib)+sizeof(ptDib^));
end;


function TDib256.GetPixelAt(X,Y: Integer): Byte;
begin
  if (X < 0) or (X >= Width) or (Y < 0) or (Y >= Height) then
    raise EWinGError.Create('Pixel coordinates out of canvas')
  else
    Result := Byte((PChar(Pixels) + X + Y * WidthBytes)^);
end;

procedure TDib256.SetPixelAt(X, Y: Integer; Value: Byte);
begin
  if (X < 0) or (X >= Width) or (Y < 0) or (Y >= Height) then
    raise EWinGError.Create('Pixel coordinates out of canvas')
  else
    (PChar(Pixels) + X + Y * WidthBytes)^ := Char(Value);
end;


function TDIB256.GetColorTable: PRGBQUAD256;
begin
   if (ptDib = nil) then
      Result := nil
   else
      Result := @(ptDib^.bmiColors);
end;




{***** Methods ******}
procedure TDIB256.LoadFromFile( const FileName: string );
var
   Stream: TFileStream;
begin
   { Open file }
   Stream := TFileStream.Create( FileName, fmOpenRead );
   try
      LoadFromStream( Stream );
   finally
      Stream.Free;
   end;
end;


procedure TDIB256.LoadFromStream( Stream: TStream );
var
   dwLen, Cnt: LongInt;
   pDt: PDIB256INFO;
   pMem, p: PChar;
   WB: Word;
begin
   pDt := GlobalAllocPtr( GMEM_FIXED, sizeof(TDIB256INFO) );

   { Read DIB structure }
   ReadBitmapInfo(Stream, pDt);

   { Calc. needed memory to put the DIB }
   WB := (pDt^.bmiHeader.biWidth+3) and $FFFC;
   dwLen  := sizeof(pDt^) + Longint(pDt^.bmiHeader.biHeight) * LongInt(WB);

   { Try realloc memory}
   pDt := GlobalReAllocPtr(pDt, dwLen, GMEM_MOVEABLE);

   { verify if DIB isn't compressed }
   case pDt^.bmiHeader.biCompression of
      BI_RGB:  Stream.ReadBuffer( (PChar(pDt)+sizeof(pDt^))^, pDt^.bmiHeader.biSizeImage);
      BI_RLE8:
      begin
         pMem := GlobalAllocPtr( GMEM_FIXED, Stream.Size-Stream.Position );
         p := pMem;
         try
            for Cnt := 1 to (Stream.Size-Stream.Position) div IOBLOCK_SIZE do
            begin
              Stream.ReadBuffer( p^, IOBLOCK_SIZE);
              if Assigned(FOnLoading) then
                FOnLoading( self,
                  ((Stream.Size-Stream.Position)*100) div ((Cnt+1)*IOBLOCK_SIZE) );
              p := p + IOBLOCK_SIZE;
            end;
            Stream.ReadBuffer( p^,
              (Stream.Size-Stream.Position) mod IOBLOCK_SIZE);
            if Assigned(FOnLoading) then
              FOnLoading( self, 100 );
            DecompressBitmap( PChar(pDt)+sizeof(pDt^), pMem, WB );
         finally
            GlobalFreePtr( pMem );
         end;
         pDt^.bmiHeader.biCompression := BI_RGB;
      end;
      else InvalidBitmap;
   end;

   { deletes last DIB, and put on the new one }
   DeleteDib;
   ptDib := pDt;

   { recalc size }
   FWidth := ptDib^.bmiHeader.biWidth;
   FHeight := ptDib^.bmiHeader.biHeight;

   { turns the image upside down }
   FlipImage;
end;


procedure TDIB256.SaveToFile( const FileName: string );
var
   Stream: TFileStream;
begin
   if ptDib = nil then
      exit;

   { Open file }
   Stream := TFileStream.Create( FileName, fmCreate );
   try
      try
         SaveToStream( Stream );
      finally
         Stream.Free;
      end;
   except
      on EWriteError do
      begin
         { Deletes file if couldn't save it all}
         DeleteFile( FileName );
         raise;
      end;
   end;
end;


procedure TDIB256.SaveToStream( Stream: TStream );
var
   dwLen, dwHeaderLen, Cnt: DWord;
   Bfh: TBitmapFileHeader;
   p: PChar;
begin
   if ptDib = nil then
      exit;

   FlipImage;
   try
      { Init. FileHeader structure }
      dwHeaderLen := sizeof(ptDib^);
      dwLen := dwHeaderLen + ptDib^.bmiHeader.biSizeImage;

      Bfh.bfType := $4D42;
      Bfh.bfSize := dwLen + sizeof(Bfh);
      Bfh.bfReserved1 := 0;
      Bfh.bfReserved2 := 0;
      Bfh.bfOffBits := dwHeaderLen + Sizeof(TBitmapFileHeader);

      { Save FileHeader }
      Stream.WriteBuffer( Bfh, sizeof(TBitmapFileHeader) );

      { Save DIB }
      p := PChar(ptDib);
      for Cnt := 0 to (dwLen div IOBLOCK_SIZE) do
      begin
        Stream.WriteBuffer( p^, IOBLOCK_SIZE);
        if Assigned(FOnSaving) then
          FOnSaving( self,
            (dwLen*100) div ((Cnt+1)*IOBLOCK_SIZE) );
        p := p + IOBLOCK_SIZE;
      end;
      Stream.WriteBuffer( p^, dwLen mod IOBLOCK_SIZE);
      if Assigned(FOnSaving) then
        FOnSaving( self, 100 );
   finally
      FlipImage;
   end;
end;


procedure TDIB256.FlipImage;
var
   Nl: Word;
   W, W4: Longint;
   pTop, pDown: Pointer;
begin
   if (ptDib = nil) or (Height < 2) then
      exit;

   { Init. vars }
   Nl := Height div 2;
   W := GetWidthBytes;
   W4 := W div 4;

   pTop := GetPixels;
   pDown := Pointer(Longint(pTop) + LongInt(GetWidthBytes) * LongInt(Height));

   asm
      push ebx
      push esi
      push edi

      mov ax, Nl
      mov esi, pTop
      mov edi, pDown
   @Loop1:
      sub edi, W
      mov ecx, W4
   @Loop2:
      mov edx, [esi]
      mov ebx, [edi]
      mov [esi], ebx
      mov [edi], edx
      add esi, 4
      add edi,4
      loop @Loop2
      sub edi, W

      dec ax
      jnz @Loop1

      pop edi
      pop esi
      pop ebx
   end;
end;




procedure TDIB256.CreateDib;
var
   dwLen, dwSizeImage: Longint;
   Cnt: Integer;
   IP: PWord;
begin
   { Delete last DIB }
   DeleteDib;

   { calc. image size and allocate memory }
   dwSizeImage := Longint( (FWidth+3) and $FFFC ) * FHeight;
   dwLen := sizeof(TDIB256INFO) + dwSizeImage;
   ptDib := GlobalAllocPtr( GMEM_FIXED, dwLen );

   { init. TBitmapInfoHeader structure}
   with ptDib^.bmiHeader do
   begin
      biSize := sizeof(TBitmapInfoHeader);
      biWidth := FWidth;
      biHeight := FHeight;
      biPlanes := 1;
      biBitCount := 8;
      biCompression := BI_RGB;
      biSizeImage := dwSizeImage;
      biXPelsPerMeter := 0;
      biYPelsPerMeter := 0;
      biClrUsed := 256;
      biClrImportant := 0;
   end;

   // init. palette
   IP := PWord(ColorTable);
   for Cnt := 0 to 255 do
   begin
     IP^ := Cnt;
     Inc(IP);
   end;
end;


procedure TDIB256.DeleteDib;
begin
   if ptDib <> nil then
      GlobalFreePtr( ptDib );

   ptDib := nil;
end;




procedure TDIB256.ReadBitmapInfo( fs: TStream; pD: PDIB256INFO );
var
   nNumColors: Integer;
   Bfh: TBitmapFileHeader;

begin
   { Read BitmapFileHeader }
   if ( fs.Read(Bfh, sizeof(Bfh)) <> sizeof(Bfh) ) then
      InvalidBitmap;

   { Se if the file has the Bitmap identify header, $4d42 = 'BM' }
   if (Bfh.bfType <> $4d42) then
      InvalidBitmap;


   with pD^ do
   begin
      { Read the BitmapInfoHeader }
      if ( fs.Read(bmiHeader, sizeof(bmiHeader)) <> sizeof(bmiHeader) ) then
         InvalidBitmap;
      if (bmiHeader.biSize <> sizeof(bmiHeader)) then
         InvalidBitmap;

      { Guarantees that the image format is valid }
      with bmiHeader do
         if (biWidth = 0) or (biHeight = 0)
         or (biBitCount <> 8) or (biPlanes <> 1) then
            InvalidBitmap;

      { }
      nNumColors := bmiHeader.biClrUsed;
      if (nNumColors = 0) then
         nNumColors := 256;

      { Read the palette }
      if nNumColors <> 0 then
         fs.ReadBuffer( bmiColors, nNumColors * sizeof(TRGBQuad) );

      { fix structure, filling optional fields }
      with bmiHeader do
      begin
         if (biSizeImage = 0) then
            biSizeImage := LongInt((biWidth+3) and $FFFC) * LongInt(biHeight);
         biClrUsed := 256;
      end;
   end;
end;


procedure TDib256.InvalidBitmap;
begin
   raise EInvalidGraphic.Create('Bitmap Image is not valid');
end;


procedure DecompressBitmap( pDest, pSource: Pointer; wWidth: Word );
var
   Cnt, c, vc: Byte;
   dif: Longint;
   pS, pD: Longint;
   Out: Boolean;
begin
   pS := Longint(pSource);
   pD := Longint(pDest);
   Out := False;

   repeat
      c := PByte(pS)^;
      Inc(pS);
      case c of
         0:
         begin
            c := PByte(pS)^;
            Inc(pS);
            case c of
               0: // end of line
               begin
                  dif := (pD - Longint(pDest)) mod wWidth;
                  if dif <> 0 then
                     pD := pD + wWidth - dif;
               end;

               1: // end of bitmap
               Out := True;

               2: // delta
               begin
                  pD := pD + PByte(pS)^;  // move right
                  Inc(pS);
                  pD := pD + PByte(pS)^*wWidth; // mode down
                  Inc(pS);
               end;

               else
               begin // absolute mode
                  for Cnt := 1 to c do
                  begin
                     PByte(pD)^ := PByte(pS)^;
                     Inc(pS);
                     Inc(pD);
                  end;
                  // word aligned
                  if ((c and 1) <> 0) then
                     Inc(pS);
               end;
            end;
         end

         else
         // encoded mode
         begin
            vc := PByte(pS)^;
            Inc(pS);
            for Cnt := 1 to c do
            begin
               PByte(pD)^ := vc;
               Inc(pD);
            end;
         end;
      end;
   until (Out = True);
end;


function CompressBitmap( pSource, pDest: Pointer; NBytes: Longint ): Longint;
begin
asm
   push ebx
	push esi
	push edi

   mov esi, pSource
   mov edi, pDest

   mov edx, NBytes      // edx - end of bitmap offset
   add edx, esi

@RLECo_loop1:
   mov al, [esi]
   cmp [esi+1], al
   jne @RLECo_menos3
   cmp [esi+2], al
   jne @RLECo_menos3

   //*** at least 3 repetitions ***
   xor ebx, ebx      // ebx - num. of repetitions (max. 255)
@RLECo_nrep:
   inc esi
   inc bl
   cmp bl, 255
   je @RLECo_colocrep
   cmp esi, edx
   je @RLECo_colocrep
   cmp [esi], al
   je @RLECo_nrep

@RLECo_colocrep:
   mov [edi], bl
   inc edi
   mov [edi], al
   inc edi
   jmp @RLECo_next


@RLECo_menos3:
   //*** no repetition ***
   mov eax, edx         // eax - max. num. of bytes until end of bitmap
   sub eax, esi

   cmp eax, 2
   jg @RLECo_menos3_1
   //*** 1 or 2 bytes until end ***
   mov al, [esi]
   xor ebx, ebx
   jmp @RLECo_nrep

@RLECo_menos3_1:
   mov ecx, 2         // ecx - num. of bytes until next repetition (min. 3)
@RLECo_proxrep:
   inc cl
   cmp cl, 254
   je @RLECo_colocbytes
   cmp ecx, eax
   je @RLECo_colocbytes
   mov al, [esi+ecx]
   cmp al, [esi+ecx+1]
   jne @RLECo_proxrep
   cmp al, [esi+ecx+2]
   jne @RLECo_proxrep

@RLECo_colocbytes:
   mov byte ptr [edi], 0
   inc edi
   mov [edi], cl
   inc edi
   mov bl, cl

@RLECo_colocbytes2:
   mov al, [esi]
   mov [edi], al
   inc esi
   inc edi
   loop @RLECo_colocbytes2

   and bl, 1                  // if num. of bytes is odd aligns to double-byte
   jz @RLECo_next
   mov byte ptr es:[edi], 0
   inc edi

@RLECo_next:
   cmp esi, edx
   jb @RLECo_loop1

   mov byte ptr es:[edi], 0
   inc edi
   mov byte ptr es:[edi], 1
   inc edi

   mov edx, edi
   xor ecx, ecx
   mov cx, word ptr [pDest]
   sub edx, ecx
   mov Result, edx

	pop edi
	pop esi
	pop ebx
end;
end;



procedure TransCopyDIBBits( pDest, pSource: Pointer; dwWidth, dwHeight: Longint;
   dwScanD, dwScanS: Longint; bTranClr: Byte );
begin
asm
      push esi
      push edi
      push ebx

      mov ecx, dwWidth
		or ecx,ecx
		jz @tcdb_nomore      // test for silly case

		mov edx, dwHeight    // EDX is line counter
		mov ah, bTranClr     // AL has transparency color

      mov esi, pSource

      mov edi, pDest

      sub dwScanD,ecx      // bias these
		sub dwScanS,ecx

		mov ebx,ecx          // save this for later

@tcdb_morelines:
      mov ecx, ebx         // ECX is pixel counter
      shr ecx,2
      jz  @tcdb_nextscan
{
; The idea here is to not branch very often so we unroll the loop by four
; and try to not branch when a whole run of pixels is either transparent
; or not transparent.
;
; There are two loops. One loop is for a run of pixels equal to the
; transparent color, the other is for runs of pixels we need to store.
;
; When we detect a "bad" pixel we jump to the same position in the
; other loop.
;
; Here is the loop we will stay in as long as we encounter a "transparent"
; pixel in the source.
}

@tcdb_same:
      mov al, [esi]
      cmp al, ah
      jne @tcdb_diff0

@tcdb_same0:
      mov al, [esi+1]
      cmp al, ah
      jne @tcdb_diff1

@tcdb_same1:
      mov al, [esi+2]
      cmp al, ah
      jne @tcdb_diff2

@tcdb_same2:
      mov al, [esi+3]
      cmp al, ah
      jne @tcdb_diff3

@tcdb_same3:
      add edi,4
      add esi,4
      dec ecx
      jnz @tcdb_same
      jz  @tcdb_nextscan

{
; Here is the loop we will stay in as long as
; we encounter a "non transparent" pixel in the source.
}

@tcdb_diff:
      mov al, [esi]
      cmp al, ah
      je @tcdb_same0

@tcdb_diff0:
      mov [edi],al
      mov al, [esi+1]
      cmp al, ah
      je @tcdb_same1

@tcdb_diff1:
      mov [edi+1],al
      mov al, [esi+2]
      cmp al, ah
      je @tcdb_same2

@tcdb_diff2:
      mov [edi+2],al
      mov al, [esi+3]
      cmp al, ah
		je @tcdb_same3

@tcdb_diff3:
		mov [edi+3],al

      add edi,4
      add esi,4
      dec ecx
      jnz @tcdb_diff
      jz @tcdb_nextscan

{
; We are at the end of a scan, check for odd leftover pixels to do
; and go to the next scan.
}

@tcdb_nextscan:
      mov ecx,ebx
      and ecx,11b
      jnz @tcdb_oddstuff
      // move on to the start of the next line

@tcdb_nextscan1:
      add esi, dwScanS
      add edi, dwScanD

      dec edx                 // line counter
		jnz @tcdb_morelines
      jz @tcdb_nomore

{
; If the width is not a multiple of 4 we will come here to clean up
; the last few pixels
}
@tcdb_oddstuff:
        inc ecx
@tcdb_oddloop:
        dec ecx
        jz  @tcdb_nextscan1
        mov al, [esi]
        inc esi
        inc edi
        cmp al, ah
        je  @tcdb_oddloop
        mov [edi-1],al
        jmp @tcdb_oddloop

@tcdb_nomore:
      pop ebx
      pop edi
      pop esi
end;
end;



procedure CopyDIBBits( pDest, pSource: Pointer; dwWidth, dwHeight: Longint;
   dwScanD, dwScanS: Longint );
begin
asm
      push es
      push esi
		push edi
      push ebx

      mov ecx, dwWidth
		or ecx,ecx
		jz @cdb_nomore       // test for silly case

		mov edx, dwHeight    // EDX is line counter
		or edx,edx
		jz @cdb_nomore       // test for silly case

      mov esi, pSource
      mov edi, pDest       // ES:[EDI] point to dest
      mov ax, ds
      mov es, ax           // es = ds

      sub dwScanD,ecx      // bias these
		sub dwScanS,ecx

      mov ebx,ecx
      shr ebx,2

      mov eax,ecx
      and eax,11b

@cdb_loop:
      mov ecx, ebx
      rep movsd
      mov ecx, eax
      rep movsb

      add esi, dwScanS
      add edi, dwScanD
      dec edx                // line counter
      jnz @cdb_loop

@cdb_nomore:
      pop ebx
      pop edi
      pop esi
      pop es
end;
end;


procedure TDib256.Copy(Dest: TDib256; SrcX, SrcY, DestX, DestY,
  CopyWidth, CopyHeight: Integer);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Width, Height, Dest.Width, Dest.Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Pixels) + SrcX + SrcY * WidthBytes;
    PDest := PChar(Dest.Pixels) + DestX + DestY * Dest.WidthBytes;
    CopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, Dest.WidthBytes,
      WidthBytes);
  end;
end;


procedure TDib256.TransCopy(Dest: TDib256; SrcX, SrcY, DestX, DestY,
  CopyWidth, CopyHeight: Integer; TransColor: Byte);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Width, Height, Dest.Width, Dest.Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Pixels) + SrcX + SrcY * WidthBytes;
    PDest := PChar(Dest.Pixels) + DestX + DestY * Dest.WidthBytes;
    TransCopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, Dest.WidthBytes,
      WidthBytes, TransColor);
  end;
end;

procedure TDib256.CopyToWinG(Dest: TWinG; SrcX, SrcY, DestX, DestY, CopyWidth,
  CopyHeight: Integer);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Width, Height, Dest.Width, Dest.Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Pixels) + SrcX + SrcY * WidthBytes;
    PDest := PChar(Dest.Pixels) + DestX + DestY * Dest.WidthBytes;
    CopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, Dest.WidthBytes,
      WidthBytes);
  end;
end;

procedure TDib256.TransCopyToWinG(Dest: TWinG; SrcX, SrcY, DestX, DestY,
  CopyWidth, CopyHeight: Integer; TransColor: Byte);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Width, Height, Dest.Width, Dest.Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Pixels) + SrcX + SrcY * WidthBytes;
    PDest := PChar(Dest.Pixels) + DestX + DestY * Dest.WidthBytes;
    TransCopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, Dest.WidthBytes,
      WidthBytes, TransColor);
  end;
end;


procedure TDib256.CopyFromWinG(Source: TWinG; SrcX, SrcY, DestX, DestY, CopyWidth,
  CopyHeight: Integer);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Source.Width, Source.Height, Width, Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True
  then begin
    PSrc := PChar(Source.Pixels) + SrcX + SrcY * Source.WidthBytes;
    PDest := PChar(Pixels) + DestX + DestY * WidthBytes;
    CopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, WidthBytes,
      Source.WidthBytes);
  end;
end;

procedure TDib256.TransCopyFromWinG(Source: TWinG; SrcX, SrcY, DestX, DestY,
  CopyWidth, CopyHeight: Integer; TransColor: Byte);
var
  PSrc, PDest: PChar;
begin
  if ClipCopyParameters(Source.Width, Source.Height, Width, Height,
    DestX, DestY, SrcX, SrcY, CopyWidth, CopyHeight) = True then
  begin
    PSrc := PChar(Source.Pixels) + SrcX + SrcY * Source.WidthBytes;
    PDest := PChar(Pixels) + DestX + DestY * WidthBytes;
    TransCopyDIBBits(PDest, PSrc, CopyWidth, CopyHeight, WidthBytes,
      Source.WidthBytes, TransColor);
  end;
end;

end.
