unit Pal256;

interface
uses Windows, SysUtils, Classes, Dialogs;

const  
   PC_RESERVED = 1;
   PC_EXPLICIT = 2;
   PC_NOCOLLAPSE = 4;

type
   TPalEntries = array [0..255] of TPALETTEENTRY;
   TLogPalette256 = record
         palVersion: Word;
         palNumEntries: Word;
         palEntry: TPalEntries;
   end;
   PPalEntries = ^TPalEntries;

   TPaletteChangeEvent = procedure (Sender: TObject; StartIndex: Integer;
      ColorCount: Integer) of object;

   TPalette256 = class(TPersistent)
   private
      FOwnHandle: Boolean;

      FHandle: HPALETTE;
      FIdentity: Boolean;

      FOnIdentityChange: TNotifyEvent;
      FOnNewPalette: TNotifyEvent;
      FOnBeforeAnimatePalette: TNotifyEvent;
      FOnAfterAnimatePalette: TPaletteChangeEvent;

      procedure SetIdentity( Value: Boolean );
      procedure SetHandle( Value: HPALETTE );
      function GetEntries: PPalEntries;

   protected
      FPalette: TLogPalette256;

   public
      property Entries: PPalEntries read GetEntries;
      property Handle: HPALETTE read FHandle write SetHandle;

      constructor Create;
      destructor Destroy; override;

      procedure Animate( StartIndex: Integer; ColorCount: Integer );
      procedure CreateHandle;
      procedure LoadFromStream( Stream: TStream );
      procedure ReadFromRGBQuad( Ptr: Pointer; NumColors: Integer );
      procedure SaveToStream( Stream: TStream );
      procedure WriteToRGBQuad( Ptr: Pointer; NumColors: Integer );

   published
      property Identity: Boolean read FIdentity write SetIdentity;

      property OnIdentityChange: TNotifyEvent read FOnIdentityChange
         write FOnIdentityChange;
      property OnNewPalette: TNotifyEvent read FOnNewPalette
         write FOnNewPalette;
      property OnBeforeAnimatePalette: TNotifyEvent read FOnBeforeAnimatePalette
         write FOnBeforeAnimatePalette;
      property OnAfterAnimatePalette: TPaletteChangeEvent
         read FOnAfterAnimatePalette write FOnAfterAnimatePalette;
   end;


implementation

constructor TPalette256.Create();
begin
   inherited Create;

{REG}

   FIdentity := True;
   FOwnHandle := True;
   FHandle := 0;

   FPalette.palVersion := $300;
   FPalette.palNumEntries := 256;
end;


destructor TPalette256.Destroy;
begin
   if (FOwnHandle = True) and (FHandle <> 0) then
      DeleteObject( FHandle );

	inherited Destroy;
end;


procedure TPalette256.SetIdentity( Value: Boolean );
begin
   if (Value = FIdentity) then
      exit;

   FIdentity := Value;

   if Assigned(FOnIdentityChange) then OnIdentityChange(Self);
end;


procedure TPalette256.SetHandle( Value: HPALETTE );
begin
   // delete current handle if mine
   if (FOwnHandle = True) and (FHandle <> 0) then
      DeleteObject( FHandle );

   // Get handle
   FOwnHandle := False;
   FHandle := Value;

   GetPaletteEntries( FHandle, 0, 256, Entries^ );

   // executes event
   if Assigned(FOnNewPalette) then OnNewPalette(Self);
end;


function TPalette256.GetEntries: PPalEntries;
begin
   Result := @(FPalette.palEntry[0]);
end;


procedure TPalette256.Animate( StartIndex: Integer; ColorCount: Integer );
begin
   if (FHandle = 0) then
      raise Exception.Create('TPalette256 has no handle to animate');

   // executes event
   if Assigned(FOnBeforeAnimatePalette) then OnBeforeAnimatePalette(Self);

   AnimatePalette( FHandle, StartIndex, ColorCount,
      @(FPalette.palEntry[StartIndex]) );

   // executes event
   if Assigned(FOnAfterAnimatePalette) then OnAfterAnimatePalette(Self,
      StartIndex, ColorCount);
end;


procedure TPalette256.CreateHandle();
var
   Cnt: Integer;
   dcDesktop: HDC;
begin
   // delete previous handle, if it's mine of course
   if (FHandle <> 0) and (FOwnHandle = True) then
      DeleteObject( FHandle );

   // define logical structure for identity palette
   if (FIdentity = True) then
   begin
      // init. static colors indexes 0 to 9 and 246 to 255 with System Colors
      // Get the static colors
      dcDesktop := GetDC(0);
      GetSystemPaletteEntries( dcDesktop, 0, 10, FPalette.palEntry[0] );
      GetSystemPaletteEntries( dcDesktop, 246, 10, FPalette.palEntry[246] );
      ReleaseDC( 0, dcDesktop );

      // init. all other colors flags to PC_RESERVED
      for Cnt := 10 to 245 do
         FPalette.palEntry[Cnt].peFlags := PC_RESERVED;
   end;

   FOwnHandle := True;
   FHandle := CreatePalette( TLogPalette(Pointer(@FPalette)^) );

   // executes event
   if Assigned(FOnNewPalette) then OnNewPalette(Self);
end;


procedure TPalette256.LoadFromStream(Stream: TStream);
begin
      Stream.Read( Entries^, sizeof(Entries^) );
end;

procedure TPalette256.SaveToStream(Stream: TStream);
begin
      Stream.Write( Entries^, sizeof(Entries^) );
end;


procedure TPalette256.ReadFromRGBQuad(Ptr: Pointer; NumColors: Integer);
var
   Cnt: Integer;
   PColor: PRGBQuad;
begin
   PColor := Ptr;
   for Cnt := 0 to NumColors-1 do
   begin
      FPalette.palEntry[Cnt].peRed := PColor^.rgbRed;
      FPalette.palEntry[Cnt].peBlue := PColor^.rgbBlue;
      FPalette.palEntry[Cnt].peGreen := PColor^.rgbGreen;
      FPalette.palEntry[Cnt].peFlags := 0;
      PColor := PRGBQuad(PChar(PColor) + sizeof(TRGBQuad));
   end;
end;



procedure TPalette256.WriteToRGBQuad(Ptr: Pointer; NumColors: Integer);
var
   Cnt: Integer;
   PColor: PRGBQuad;
begin
   PColor := Ptr;
   for Cnt := 0 to NumColors-1 do
   begin
      PColor^.rgbRed := FPalette.palEntry[Cnt].peRed;
      PColor^.rgbBlue := FPalette.palEntry[Cnt].peBlue;
      PColor^.rgbGreen := FPalette.palEntry[Cnt].peGreen;
      PColor^.rgbReserved := 0;
      PColor := PRGBQuad(PChar(PColor) + sizeof(TRGBQuad));
   end;
end;


end.

