////////////////////////////////////////////////////////////////////////////////
// Main unit of the HITThreadDemo project
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit HITThreadMain;

interface
{$A+,H+}

uses
  Classes, Forms, Controls, StdCtrls, Integers, HITComputationThreads;

type
  TFactorThread = class(THITComputationThread)
    fListBox: TListBox;
    fProgress: Longint;
    fF, fN: TInteger;
    procedure EditAborted;
    procedure EditException;
    procedure EditFactor;
    procedure EditFailed;
    procedure EditProgress;
    procedure InitEditor;
  public
    constructor Create(ListBox: TListBox; F, N: TInteger;
                       DoOnTerminate: TNotifyEvent);
    procedure Computation; override;
  end;

  TThreadComputingForm = class(TForm)
    Memo1: TMemo;
    ListBox1: TListBox;
    ListBox2: TListBox;
    ListBox3: TListBox;
    ListBox4: TListBox;
    AbortBtn1: TButton;
    AbortBtn2: TButton;
    AbortBtn3: TButton;
    AbortBtn4: TButton;
    StartBtn: TButton;
    CloseBtn: TButton;
    procedure AbortBtn1Click(Sender: TObject);
    procedure AbortBtn2Click(Sender: TObject);
    procedure AbortBtn3Click(Sender: TObject);
    procedure AbortBtn4Click(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure StartBtnClick(Sender: TObject);
  private
    fN, fF1, fF2, fF3, fF4: TInteger;
    fNStr1, fNStr2: string;
    fComp1, fComp2, fComp3, fComp4: TFactorThread;
    fRunningThreads: Integer;
    fInitialized: Boolean;
    procedure AllThreadsDone(Sender: TObject);
    procedure GetNewIntegers;
    procedure Thread1Done(Sender: TObject);
    procedure Thread2Done(Sender: TObject);
    procedure Thread3Done(Sender: TObject);
    procedure Thread4Done(Sender: TObject);
  end;

var
  ThreadComputingForm: TThreadComputingForm;


implementation
//==============================================================================
{$C-,I-,O+,Q-,R-,W-}
//==============================================================================
uses Graphics, SysUtils, Integers_Primes;
{$R *.DFM}

////////////////////////////////////////////////////////////////////////////////
// TFactorThread class
////////////////////////////////////////////////////////////////////////////////

constructor TFactorThread.Create(ListBox: TListBox; F, N: TInteger;
                                 DoOnTerminate: TNotifyEvent);
begin
  //----- Just copy pointers
  fListBox := ListBox;
  fF := F;
  fN := N;
  inherited Create(true); //Suspended, need Resume to start
  FreeOnTerminate := true;
  OnTerminate := DoOnTerminate;
  Synchronize(InitEditor);
end;

procedure TFactorThread.EditAborted;
begin
  Beep;
  with fListBox, Items do
  begin
    Color := clAqua;
    BeginUpdate;
    Add('');
    Add(' THREAD ABORTED');
    EndUpdate;
  end;
end;

procedure TFactorThread.EditException;
begin
  Beep;
  with fListBox, Items do
  begin
    Color := clRed;
    Clear;
    BeginUpdate;
    Add('');
    Add(' EXCEPTION:');
    Add(' ' + ExceptName);
    Add(' ' + ExceptMsg);
    EndUpdate;
  end;
end;

procedure TFactorThread.EditFailed;
begin
  Beep;
  with fListBox, Items do
  begin
    Color := clLime;
    BeginUpdate;
    Add('');
    Add(' Failed');
    EndUpdate;
  end;
end;

procedure TFactorThread.EditFactor;
begin
  Beep;
  with fListBox.Items do
  begin
    BeginUpdate;
    Add('');
    Add(' Factor found:');
    Add(' ' + IStr(fF,10,0));
    EndUpdate;
  end;
end;

procedure TFactorThread.EditProgress;
begin
  with fListBox, Items do
  begin
    BeginUpdate;
    Items[Items.Count-1] := ' ' + IntToStr(fProgress);
    EndUpdate;
  end;
end;

procedure TFactorThread.InitEditor;
begin
  with fListBox, Items do
  begin
    Color := clWhite;
    Clear;
    BeginUpdate;
    Add('');
    Add(' Seed:');
    Add(IStr(fF,10,0));
    Add('');
    Add(' Loop counter:');
    Add(' 0');
    EndUpdate;
  end;
end;

//==============================================================================
// Try to factor N
// -> N, composite integer that has no "small" factors
// -> F, random integer such that 0 < F < N
// <- F, factor of N if the factorization succeeded
//    F = 1 or F = N if the factorization failed
//==============================================================================
procedure TFactorThread.Computation;
  const
    MAXLOOP = Longword(8192);
    MAXPROD = Longword(32); //Must be much smaller than the factors of N
  var
    X, Y, Z    : TInteger;
    i, imax, j : Longword;
    Progress   : Longint;
begin
  X := nil;
  Y := nil;
  Z := nil;
  try
    //----- Create integers equal to 0
    X := TInteger.Create;
    Y := TInteger.Create;
    Z := TInteger.Create;

    //----- Raise an exception (just for demo, you can suppress this line)
    // When using this line with the DELPHI debugger on, disable the
    // Tools/Debugger options/Language Exceptions/Stop on DELPHI Exceptions
    // check box
//  if IMod(fF,8) = 3 then HITError('Just for demo!');

    //----- Initialization
    ISwp(X,fF); //X := F and F := 0
    fF.Set1;
    imax := 1;
    Progress := 0;
    repeat
      ISet(Y,X);
      for i := 1 to imax * MAXPROD do
      begin
        Inc(Progress);
        if (Progress and 1023) = 0 then
        begin

          // Since EditProgress uses a VCL component (TListBox) it should never
          // be called directly by the thread. EditProgress must be called with
          // the Synchronize method in order to avoid multi-thread conflicts.
          // The parameters Progress is copied to a field so that the main VCL
          // thread can access it when executing EditProgress.

          fProgress := Progress shr 10;
          Synchronize(EditProgress);
        end;
        //----- X := X + 1 mod N
        ISqr(X);
        IInc(X);
        IMod(X,fN);
        //----- Check if should terminate
        if Terminated then
        begin
          Synchronize(EditAborted);
          Exit;
        end;
      end;
      for i := 1 to imax do
      begin
        //----- When starting j loop, we always have D = 1
        for j := 1 to MAXPROD do
        begin
          Inc(Progress);
          if (Progress and 1023) = 0 then
          begin
            fProgress := Progress shr 10;
            Synchronize(EditProgress);
          end;
          //----- X := X + 1 mod N
          ISqr(X);
          IInc(X);
          IMod(X,fN);
          //----- D := (X - Y) * D mod N
          ISet(Z,X);
          ISub(Z,Y); //Z may be negative. Don't mind.
          IMul(fF,Z);
          IMod(fF,fN);
          //----- Check if should terminate
          if Terminated then
          begin
            Synchronize(EditAborted);
            Exit;
          end;
        end;
        //----- If GCD(F,N) <> 1 then exit
        //      If GCD(F,N) = N, one must retry with a different
        //      initial value for F (input)
        if not IGCD(fF,fN) then
        begin
          if IEqu(fF,fN) then Synchronize(EditFailed)
                         else Synchronize(EditFactor);
          Exit;
        end;
      end;
      //----- Check if should terminate
      if Terminated then
      begin
        Synchronize(EditAborted);
        Exit;
      end;
      Inc(imax,imax);
    until imax > MAXLOOP;
    //----- Failed factorization
    Synchronize(EditFailed);
  finally
    //----- Free the integers
    X.Free;
    Y.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TThreadComputingForm class
////////////////////////////////////////////////////////////////////////////////

procedure TThreadComputingForm.AbortBtn1Click(Sender: TObject);
begin
  AbortBtn1.Enabled := False;
  fComp1.Terminate;
end;

procedure TThreadComputingForm.AbortBtn2Click(Sender: TObject);
begin
  AbortBtn2.Enabled := False;
  fComp2.Terminate;
end;

procedure TThreadComputingForm.AbortBtn3Click(Sender: TObject);
begin
  AbortBtn3.Enabled := False;
  fComp3.Terminate;
end;

procedure TThreadComputingForm.AbortBtn4Click(Sender: TObject);
begin
  AbortBtn4.Enabled := False;
  fComp4.Terminate;
end;

procedure TThreadComputingForm.AllThreadsDone(Sender: TObject);
begin
  fInitialized := False;
  StartBtn.Enabled := True;
  CloseBtn.Enabled := True;
end;

procedure TThreadComputingForm.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

procedure TThreadComputingForm.FormCreate(Sender: TObject);
begin
  fN := TInteger.Create;
  fF1 := TInteger.Create;
  fF2 := TInteger.Create;
  fF3 := TInteger.Create;
  fF4 := TInteger.Create;
end;

procedure TThreadComputingForm.FormDestroy(Sender: TObject);
begin
  fN.Free;
  fF1.Free;
  fF2.Free;
  fF3.Free;
  fF4.Free;
end;

procedure TThreadComputingForm.GetNewIntegers;
begin
  if not fInitialized then
  begin
    ITimeSeed;
    //----- Generate N so that it is the product of "not too small" primes
    IRnd(fF1,32);
    fF1.Odd := true;
    while not ISPP(fF1,2) do IAdd(fF1,2);
    IRnd(fF2,32);
    fF2.Odd := true;
    while not ISPP(fF2,2) do IAdd(fF2,2);
    IRnd(fF3,32);
    fF3.Odd := true;
    while not ISPP(fF3,2) do IAdd(fF3,2);
    IRnd(fF4,110);
    fF4.Odd := true;
    while not ISPP(fF4,2) do IAdd(fF4,2);
    ISet(fN,fF1);
    IMul(fN,fF2);
    IMul(fN,fF3);
    IMul(fN,fF4);
    //----- Make strings
    fNStr1 := IStr(fN,10,0)  + ' =';
    fNStr2 := IStr(fF1,10,0) + ' *' +
              IStr(fF2,10,0) + ' *' +
              IStr(fF3,10,0) + ' *' +
              IStr(fF4,10,0);
    //----- Random seeds
    IRnd(fF1,32,false);
    IRnd(fF2,32,false);
    IRnd(fF3,32,false);
    IRnd(fF4,32,false);
    fInitialized := True;
  end;
end;

procedure TThreadComputingForm.StartBtnClick(Sender: TObject);
begin
  StartBtn.Enabled := False;
  CloseBtn.Enabled := False;
  //----- Get a new integer to factor and new seeds
  GetNewIntegers;
  //----- Update Memo1
  Memo1.Color := clWhite;
  Memo1.Lines.BeginUpdate;
  Memo1.Lines[2] := fNStr1;
  Memo1.Lines[3] := fNStr2;
  Memo1.Lines.EndUpdate;
  //----- Create and launch threads
  fRunningThreads := 4;
  fComp1 := TFactorThread.Create(ListBox1,fF1,fN,Thread1Done);
  fComp2 := TFactorThread.Create(ListBox2,fF2,fN,Thread2Done);
  fComp3 := TFactorThread.Create(ListBox3,fF3,fN,Thread3Done);
  fComp4 := TFactorThread.Create(ListBox4,fF4,fN,Thread4Done);
  fComp1.Resume;
  fComp2.Resume;
  fComp3.Resume;
  fComp4.Resume;
  //----- Allow to abort threads
  AbortBtn1.Enabled := True;
  AbortBtn2.Enabled := True;
  AbortBtn3.Enabled := True;
  AbortBtn4.Enabled := True;
end;

procedure TThreadComputingForm.Thread1Done(Sender: TObject);
begin
  AbortBtn1.Enabled := False;
  if fComp1.ExceptRaised then fComp1.EditException;
  Dec(fRunningThreads);
  if fRunningThreads = 0 then AllThreadsDone(Sender);
end;

procedure TThreadComputingForm.Thread2Done(Sender: TObject);
begin
  AbortBtn2.Enabled := False;
  if fComp2.ExceptRaised then fComp2.EditException;
  Dec(fRunningThreads);
  if fRunningThreads = 0 then AllThreadsDone(Sender);
end;

procedure TThreadComputingForm.Thread3Done(Sender: TObject);
begin
  AbortBtn3.Enabled := False;
  if fComp3.ExceptRaised then fComp3.EditException;
  Dec(fRunningThreads);
  if fRunningThreads = 0 then AllThreadsDone(Sender);
end;

procedure TThreadComputingForm.Thread4Done(Sender: TObject);
begin
  AbortBtn4.Enabled := False;
  if fComp4.ExceptRaised then fComp4.EditException;
  Dec(fRunningThreads);
  if fRunningThreads = 0 then AllThreadsDone(Sender);
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

