////////////////////////////////////////////////////////////////////////////////
// Gaussian Integers (Definitions and basic operations)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GaussIntegers;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers, Classes;

//==============================================================================
// The Gaussian integers Z[i] are the complex numbers whose real and imaginary
// parts are integers. The symbol i is such that i = -1.
// Z[i] is a commutative ring.
//------------------------------------------------------------------------------
// An element z of Z[i] is denoted z = a + ib or sometimes (a,b).
// The conjugate of z = a + ib is ~z = a - ib.
// The norm of z is N(z) = z*~z = a + b.
// A Gaussian integer is odd, if and only if its norm is odd.
// A Gaussian integer of which the norm equals 1 is called a "unit".
// The 4 units of Z[i] are 1, -1, i and -i.
// The GCD of 2 numbers is unique (up to a multiplication by a unit).
// The prime numbers of Z[i] are (up to a multiplication by a unit):
// 1) the prime numbers congruent to 3 modulo 4 in N (natural numbers);
// 2) a + ib, such that a + b is prime in N.
//==============================================================================

type
  TGaussInt = class(THITObject)
  private
    // Fields
    fR: TInteger; //Real part
    fI: TInteger; //Imaginary part
  public
    // Constructors and destructor
    constructor Create; overload;
    constructor Create(R, I: Int64); overload;
    constructor Create(R, I: TInteger); overload;
    constructor Create(Z: TGaussInt); overload;
    destructor  Destroy; override;
    // Methods
    function  Clone: TGaussInt;
    procedure Conjugate; //(R,I) := (R,-I)
    procedure Copy(Z: TGaussInt);
    function  Is0: Boolean; //Z = (0,0) ?
    function  Is1: Boolean; //Z = (1,0) ?
    function  IsI: Boolean; //Z = (0,1) ?
    function  IsUnit: Boolean; //R+I = 1 ? (i.e., Z is a unit ?)
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure MulByI;
    procedure MulByMinusI;
    procedure Negate;
    procedure Norm(N: TInteger);
    function  Odd: Boolean;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure Set0; //Z := (0,0)
    procedure Set1; //Z := (1,0)
    procedure SetI(I: Int64); overload;    //Z := (unmodified,I)
    procedure SetI(I: TInteger); overload; //Z := (unmodified,I)
    procedure SetR(R: Int64); overload;    //Z := (R,unmodified)
    procedure SetR(R: TInteger); overload; //Z := (R,unmodified)
    // Properties
    property  Im: TInteger read fI;
    property  Re: TInteger read fR;
  end;

//===== Basic operations
procedure ZAdd(A, B: TGaussInt);
procedure ZDiv(A, B: TGaussInt);
function  ZEqu(A, B: TGaussInt): Boolean;
procedure ZGCD(A, B: TGaussInt);
function  ZGCDU(A, B: TGaussInt): Boolean;
function  ZInvMod(A, B: TGaussInt): Boolean;
procedure ZMod(A, B: TGaussInt);
procedure ZModDiv(A, B: TGaussInt);
procedure ZMul(A, B: TGaussInt);
procedure ZPow(A: TGaussInt; E: TInteger);
function  ZPowMod(A: TGaussInt; E: Int64; B: TGaussInt): Boolean; overload;
function  ZPowMod(A: TGaussInt; E: TInteger; B: TGaussInt): Boolean; overload;
procedure ZSqr(A: TGaussInt);
procedure ZSub(A, B: TGaussInt);
procedure ZSet(A, B: TGaussInt); overload;
procedure ZSet(A: TGaussInt; R, I: Int64); overload;
procedure ZSet(A: TGaussInt; R, I: TInteger); overload;
procedure ZSwp(A, B: TGaussInt);


implementation
//==============================================================================
uses Windows, SysUtils;

////////////////////////////////////////////////////////////////////////////////
// ROUTINES
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Local
// Compute centered remainders A.fR mod N and A.fI mod N
// Such remainders are called "centered" (around 0) because we choose, between
// R and N-R, the one having the smallest absolute value.
//------------------------------------------------------------------------------
procedure CenteredRemainders(A: TGaussInt; N: TInteger);
  var H : TInteger;
begin
  if N.Neg then HITRaiseInvalidArg('CenteredRemainders','N < 0');

  H := N.Clone;
  with A do
  try
    //----- H := N/2
    IShr(H);
    //----- Real part
    IMod(fR,N);
    if ICmpAbs(fR,H) > 0 then ISub(fR,N);
    //----- Imaginary part
    IMod(fI,N);
    if ICmpAbs(fI,H) > 0 then ISub(fI,N);
  finally
    H.Free;
  end;
end;

//==============================================================================
// Global
// Addition
// A := A + B
//==============================================================================
procedure ZAdd(A, B: TGaussInt);
begin
  IAdd(A.fR,B.fR);
  IAdd(A.fI,B.fI);
end;

//==============================================================================
// Global
// Division
// A := A div B
//==============================================================================
procedure ZDiv(A, B: TGaussInt);
  var
    N : TInteger;
    R : TGaussInt;
begin
  //----- Check
  if B.Is0 then HITRaiseDivByZero('ZDiv');

  if A = B then
  begin
    A.Set1;
    Exit;
  end;

  N := nil;
  R := nil;
  try
    N := TInteger.Create;
    R := A.Clone;
    //----- R := R * ~B
    B.Conjugate;
    ZMul(R,B);
    B.Conjugate;
    //----- R := (Rr/N(B), Ri/N(B))
    B.Norm(N);
    CenteredRemainders(R,N);
    //----- R := R*B/N(B)
    ZMul(R,B);
    IDiv(R.fR,N);
    IDiv(R.fI,N);
    //----- A := (A-R) * ~B/N(B)
    ZSub(A,R);
    B.Conjugate;
    ZMul(A,B);
    B.Conjugate;
    IDiv(A.fR,N);
    IDiv(A.fI,N);
  finally
    N.Free;
    R.Free;
  end;
end;

//==============================================================================
// Global
// Result := A = B
//==============================================================================
function ZEqu(A, B: TGaussInt): Boolean;
begin
  Result := IEqu(A.fR,B.fR) and IEqu(A.fI,B.fI)
end;

//==============================================================================
// Global
// GCD
// A := GCD(A,B)
// The GCD of 2 Gaussian integers is unique (up to a multiplication by a unit).
//==============================================================================
procedure ZGCD(A, B: TGaussInt);
  var T : TGaussInt;
begin
  if (A = B) or ZEqu(A,B) then Exit;

  T := B.Clone;
  try
    while not T.Is0 do
    begin
      ZMod(A,T);
      ZSwp(A,T);
    end;
  finally
    T.Free;
  end;
end;

//==============================================================================
// Global
// GCD
// Return TRUE if and only if GCD(A,B) is a unit of Z[i]
//==============================================================================
function ZGCDU(A, B: TGaussInt): Boolean;
  var S : TGaussInt;
begin
  S := A.Clone;
  try
    ZGCD(S,B);
    Result := S.IsUnit;
  finally
    S.Free;
  end;
end;

//==============================================================================
// A := 1/A mod B (if exists)
// Result := TRUE if inverse exists, FALSE otherwise
//==============================================================================
function ZInvMod(A, B: TGaussInt): Boolean;
  var R0, R1, R2, U0, U1, U2 : TGaussInt;
begin
  //----- Check
  if A = B then HITRaiseDuplicatePtr('ZInvMod','A,B');
  if B.Is0 then HITRaiseDivByZero('ZInvMod');

  R0 := nil;
  R1 := nil;
  R2 := nil;
  U0 := nil;
  U1 := nil;
  U2 := nil;
  try
    R0 := TGaussInt.Create;
    R1 := A.Clone;
    R2 := B.Clone;
    U0 := TGaussInt.Create;
    U1 := TGaussInt.Create;
    U2 := TGaussInt.Create;
    U1.Set1;
    repeat
      ZSet(R0,R1);
      ZSet(U0,R2);
      ZModDiv(R0,U0);
      U0.Negate;
      ZMul(U0,U2);
      ZAdd(U0,U1);
      ZSet(R1,R2);
      ZSet(U1,U2);
      ZSet(R2,R0);
      ZSet(U2,U0);
    until R0.Is0;
    //
    // Here, R1 = GCD(A,B) and U1 = 1/A mod B
    //
    Result := R1.IsUnit;
    if Result then //Do not change A if not invertible
    begin
      ZDiv(U1,R1); //So that (A*inv(A) mod B) = 1, and not only any unit
      ZSwp(A,U1);
    end;
    //
    // Here it possible to return the GCD in A in case Result = false
    //
  finally
    R0.Free;
    R1.Free;
    R2.Free;
    U0.Free;
    U1.Free;
    U2.Free;
  end;
end;

//==============================================================================
// Global
// Division
// A := A mod B such that Norm(A) < Norm(B)
//==============================================================================
procedure ZMod(A, B: TGaussInt);
  var N : TInteger;
begin
  //----- Check
  if B.Is0 then HITRaiseDivByZero('ZMod');

  if A = B then
  begin
    A.Set0;
    Exit;
  end;

  N := TInteger.Create;
  try
    //----- A := A * ~B
    B.Conjugate;
    ZMul(A,B);
    B.Conjugate;
    //----- A := (Ar/N(B),Ai/N(B))
    B.Norm(N);
    CenteredRemainders(A,N);
    //----- A := A*B/N(B)
    ZMul(A,B);
    IDiv(A.fR,N);
    IDiv(A.fI,N);
  finally
    N.Free;
  end;
end;

//==============================================================================
// Global
// Division
// A := A rem B  and  B := A div B
//==============================================================================
procedure ZModDiv(A, B: TGaussInt);
  var
    N : TInteger;
    T : TGaussInt;
begin
  //----- Check
  if A = B then HITRaiseDuplicatePtr('ZModDiv','A,B');
  if B.Is0 then HITRaiseDivByZero('ZModDiv');

  N := nil;
  T := A.Clone;
  try
    N := TInteger.Create;
    //----- A := A * ~B
    B.Conjugate;
    ZMul(A,B);
    B.Conjugate;
    //----- A := (Ar/N(B), Ai/N(B))
    B.Norm(N);
    CenteredRemainders(A,N);
    //----- A := A*B/N(B)
    ZMul(A,B);
    IDiv(A.fR,N);
    IDiv(A.fI,N);
    //----- B :=  (T-A)*~B/N(B)
    ZSub(T,A);
    B.Conjugate;
    ZMul(B,T);
    IDiv(B.fR,N);
    IDiv(B.fI,N);
  finally
    N.Free;
    T.Free;
  end;
end;

//==============================================================================
// Global
// Multiplication
// A := A * B = (Ar*Br - Ai*Bi, Ar*Bi + Ai*Br)
// Intermediate calculations:
//  U = Ar*Br
//  V = Ai*Bi
//  W = (Ar+Ai)*(Br+Bi)
// Results:
//  Ar := U - V
//  Ai := W - U - V
// We only do 3 multiplications instead of 4 (There are 1 addition and 2
// subtractions more, but their running times are insignificant compared to the
// one of a multiplication)
//==============================================================================
procedure ZMul(A, B: TGaussInt);
  var S, T : TInteger;
begin
  //----- This is not only faster, this is indispensable!
  if A = B then ZSqr(A)
  else
  with A do
  begin
    S := nil;
    T := B.fR.Clone;
    try
      S := fI.Clone;
      IAdd(fI,fR);
      IAdd(T,B.fI);
      IMul(fI,T);     //W := (Ar + Ai) * (Br + Bi)
      IMul(fR,B.fR);  //U  := Ar * Br
      IMul(S,B.fI);   //V  := Ai * Bi
      ISub(fI,fR);
      ISub(fI,S);     //Ai := W - U - V
      ISub(fR,S);     //Ar := U - V
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global
// Exponentiation
// A := A^E
//==============================================================================
procedure ZPow(A: TGaussInt; E: TInteger);
  var
    T : TGaussInt;
    i : Longint;
begin
  if E.Neg then HITError('ZPow'#13#10'Negative exponent');

  if E.Is0 then A.Set1
  else
  if not E.Is1 then
  begin
    T := A.Clone;
    try
      //----- Repeated squaring exponentiation
      for i := E.BitSize-2 downto 0 do
      begin
        ZSqr(A);
        if E.Bit[i] then ZMul(A,T);
      end;
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// Exponentiation modulo B
// A := A^E mod B
// Result := TRUE iff (E >= 0) or (A invertible mod B)
//==============================================================================
function ZPowMod(A: TGaussInt; E: Int64; B: TGaussInt): Boolean;
  var
    T : TGaussInt;
    i : Longint;
begin
  //----- Check
  if A = B then HITRaiseDuplicatePtr('ZPowMod','A,B');
  if B.Is0 then HITRaiseInvalidArg('ZPowMod','B = 0');

  Result := true;
  if B.Is1 then
  begin
    A.Set0;
    if E < 0 then Result := false;
  end
  else
  if E = 0 then A.Set1
  else
  begin
    //----- Reduce modulo B
    ZMod(A,B);
    if A.Is1 then Exit
    else
    if A.Is0 then
    begin
      if E < 0 then Result := false;
      Exit;
    end;
    //
    // Here A <> 1 and |E| <> 0
    //
    if E < 0 then
    begin
      Result := ZInvMod(A,B);
      if not Result then Exit;
      E := -E;
    end;
    //
    // Here, E cannot be negative except if equal to MinInt64
    //
    if (E > 1) or (E = MinInt64) then
    begin
      T := A.Clone;
      try
        //----- Repeated squaring exponentiation
        for i := IBitSize64(E)-2 downto 0 do
        begin
          ZSqr(A);
          ZMod(A,B);
          if IBit64(E,i) then
          begin
            ZMul(A,T);
            ZMod(A,B);
          end;
        end;
      finally
        T.Free;
      end;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// Exponentiation modulo B
// A := A^E mod B
// Result := TRUE iff (E >= 0) or (A invertible mod B)
//==============================================================================
function ZPowMod(A: TGaussInt; E: TInteger; B: TGaussInt): Boolean;
  var
    T           : TGaussInt;
    i, EBitSize : Longint;
begin
  //----- Check
  if A = B then HITRaiseDuplicatePtr('ZPowMod','A,B');
  if B.Is0 then HITRaiseInvalidArg('ZPowMod','B = 0');

  Result := true;
  if B.Is1 then
  begin
    A.Set0;
    if E.Neg then Result := false;
  end
  else
  if E.Is0 then A.Set1
  else
  begin
    //----- Reduce modulo B
    ZMod(A,B);
    if A.Is1 then Exit
    else
    if A.Is0 then
    begin
      if E.Neg then Result := false;
      Exit;
    end;
    //
    // Here, A <> 1 and |E| <> 0
    //
    if E.Neg then
    begin
      Result := ZInvMod(A,B);
      if not Result then Exit;
    end;
    EBitSize := E.BitSize;
    if EBitSize > 1 then
    begin
      T := A.Clone;
      try
        //----- Repeated squaring exponentiation
        for i := EBitSize-2 downto 0 do
        begin
          ZSqr(A);
          ZMod(A,B);
          if E.Bit[i] then
          begin
            ZMul(A,T);
            ZMod(A,B);
          end;
        end;
      finally
        T.Free;
      end;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure ZSet(A, B: TGaussInt);
begin
  if A <> B then
  begin
    ISet(A.fR,B.fR);
    ISet(A.fI,B.fI);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure ZSet(A: TGaussInt; R, I: Int64);
begin
  ISet(A.fR,R);
  ISet(A.fI,I);
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure ZSet(A: TGaussInt; R, I: TInteger);
begin
  ISet(A.fR,R);
  ISet(A.fI,I);
end;

//==============================================================================
// Global
// Square
// A := A = (r-i, 2ri)
//==============================================================================
procedure ZSqr(A: TGaussInt);
  var S, T : TInteger;
begin
  S := nil;
  T := nil;
  with A do
  try
    S := fR.Clone;
    T := fI.Clone;
    IMul(fI,fR);
    IShl(fI);
    IAdd(fR,T);
    ISub(S,T);
    IMul(fR,S);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Global
// Subtraction
// A := A - B
//==============================================================================
procedure ZSub(A, B: TGaussInt);
begin
  ISub(A.fR,B.fR);
  ISub(A.fI,B.fI);
end;

//==============================================================================
// Global
// Swap the values of A and B
//==============================================================================
procedure ZSwp(A, B: TGaussInt);
begin
  ISwp(A.fR,B.fR);
  ISwp(A.fI,B.fI);
end;

////////////////////////////////////////////////////////////////////////////////
// TGaussInt class
////////////////////////////////////////////////////////////////////////////////

//----- Used by "files" methods
type
  TGaussHeaderFile = array[0..20] of AnsiChar;
const
  GaussHeaderFile : TGaussHeaderFile = 'HIT Gaussian Integer ';

//==============================================================================
// Public
// Create and set Result equal to Self
//==============================================================================
function TGaussInt.Clone: TGaussInt;
begin
  Result := TGaussInt.Create(Self);
end;

//==============================================================================
// Public
// Conjugate the number
//==============================================================================
procedure TGaussInt.Conjugate;
begin
  fI.Negate;
end;

//==============================================================================
// Public
// Self := Z
//==============================================================================
procedure TGaussInt.Copy(Z: TGaussInt);
begin
  if Self <> Z then
  begin
    ISet(fR,Z.fR);
    ISet(fI,Z.fI);
  end;
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TGaussInt
//==============================================================================
constructor TGaussInt.Create(R, I: Int64);
begin
  inherited Create;
  fR := TInteger.Create(R);
  fI := TInteger.Create(I);
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TGaussInt
//==============================================================================
constructor TGaussInt.Create(R, I: TInteger);
begin
  inherited Create;
  fR := R.Clone;
  fI := I.Clone;
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TGaussInt
//==============================================================================
constructor TGaussInt.Create(Z: TGaussInt);
begin
  inherited Create;
  fR := Z.fR.Clone;
  fI := Z.fI.Clone;
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TGaussInt
//==============================================================================
constructor TGaussInt.Create;
begin
  inherited Create;
  fR := TInteger.Create;
  fI := TInteger.Create;
end;

//==============================================================================
// Public
// Destroy the instance
//==============================================================================
destructor TGaussInt.Destroy;
begin
  fR.Free;
  fI.Free;
  inherited Destroy;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is equal to (0,0)
//==============================================================================
function TGaussInt.Is0: Boolean;
begin
  Result := fR.Is0 and fI.Is0;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is equal to (1,0)
//==============================================================================
function TGaussInt.Is1: Boolean;
begin
  Result := fR.Is1 and fI.Is0;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is equal to (0,1)
//==============================================================================
function TGaussInt.IsI: Boolean;
begin
  Result := fR.Is0 and fI.Is1;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is a unit of Z[i]
// Units = {(1,0), (-1,0), (0,1), (0,-1)}
//==============================================================================
function TGaussInt.IsUnit: Boolean;
begin
  Result := (fR.Is0 and fI.Is1Abs) or (fR.Is1Abs and fI.Is0);
end;

//==============================================================================
// Public
// Load the number from a "HIT Gaussian Integer" file
//==============================================================================
procedure TGaussInt.LoadFromFile(const FileName: string);
  const FLAGS = fmOpenRead or fmShareDenyWrite;
  var
    F : TFileStream;
    H : TGaussHeaderFile;
begin
  F := TFileStream.Create(FileName,FLAGS);
  try
    //----- Header
    F.ReadBuffer(H,SizeOf(H));
    if (H <> GaussHeaderFile) then
      HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
    //----- Values
    LoadFromStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Load the number from a stream
//==============================================================================
procedure TGaussInt.LoadFromStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fR.LoadFromStream(Stream);
    fI.LoadFromStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Set Self equal to Self * i
//==============================================================================
procedure TGaussInt.MulByI;
begin
  ISwp(fR,fI);
  fR.Negate;
end;

//==============================================================================
// Public
// Set Self equal to Self * -i
//==============================================================================
procedure TGaussInt.MulByMinusI;
begin
  ISwp(fR,fI);
  fI.Negate;
end;

//==============================================================================
// Public
// Negate the number
//==============================================================================
procedure TGaussInt.Negate;
begin
  fR.Negate;
  fI.Negate;
end;

//==============================================================================
// Public
// Calculate the norm of the number
// N := R + I
//==============================================================================
procedure TGaussInt.Norm(N: TInteger);
  var T : TInteger;
begin
  T := fI.Clone;
  try
    ISet(N,fR);
    ISqr(N);
    ISqr(T);
    IAdd(N,T);
  finally
    T.Free;
  end;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is odd
// A Gaussian integer is odd iff its norm is odd, but its norm is odd iff the
// real and imaginary parts have opposite parities.
// An even (non odd) Gaussian integer is always divisible by Z = 1 + i.
//==============================================================================
function TGaussInt.Odd: Boolean;
begin
  Result := fR.Odd xor fI.Odd;
end;

//==============================================================================
// Public
// Save the number to a "HIT Gaussian Integer " file
//==============================================================================
procedure TGaussInt.SaveToFile(const FileName: string);
  const FLAGS = fmCreate or fmShareExclusive;
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,FLAGS);
  try
    //----- Header
    F.WriteBuffer(GaussHeaderFile,SizeOf(TGaussHeaderFile));
    //----- Values
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Save the number to a stream
//==============================================================================
procedure TGaussInt.SaveToStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fR.SaveToStream(Stream);
    fI.SaveToStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Self := (0,0)
//==============================================================================
procedure TGaussInt.Set0;
begin
  fR.Set0;
  fI.Set0;
end;

//==============================================================================
// Public
// Self := (1,0)
//==============================================================================
procedure TGaussInt.Set1;
begin
  fR.Set1;
  fI.Set0;
end;

//==============================================================================
// Public
// Self.Im := I
//==============================================================================
procedure TGaussInt.SetI(I: Int64);
begin
  ISet(fI,I);
end;

//==============================================================================
// Public
// Self.Im := I
//==============================================================================
procedure TGaussInt.SetI(I: TInteger);
begin
  ISet(fI,I);
end;

//==============================================================================
// Public
// Self.Re := R
//==============================================================================
procedure TGaussInt.SetR(R: Int64);
begin
  ISet(fR,R);
end;

//==============================================================================
// Public
// Self.Re := R
//==============================================================================
procedure TGaussInt.SetR(R: TInteger);
begin
  ISet(fR,R);
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

