////////////////////////////////////////////////////////////////////////////////
// One-Time Pad Dependent RSA Cryptosystem
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
//------------------------------------------------------------------------------
// For theoretical explanations see
// "HD-RSA: Hybrid Dependent RSA, a New Public-Key Encryption Scheme"
// by David Pointcheval, downloadable from
// ftp://ftp.ens.fr/pub/dmi/users/pointche/Reports/1999_HDRSA.ps.Z
////////////////////////////////////////////////////////////////////////////////
unit OneTimePadDRSA;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;


function  OTPDRSADecrypt(M, A,B,H, N,D: TInteger): Boolean;
function  OTPDRSADecryptCRT(M, A,B,H, N,DP,DQ,UP,UQ,P,Q: TInteger): Boolean;
procedure OTPDRSAEncrypt(A,B,H, M, N: TInteger);
procedure OTPDRSAKeys(N,D: TInteger; NSize: Longint);
procedure OTPDRSAKeysCRT(N,DP,DQ,UP,UQ,P,Q: TInteger; NSize: Longint);

function  OTPDRSATest(NSize: Longint; var te, td: Extended): Boolean;
function  OTPDRSATestCRT(NSize: Longint; var te, td: Extended): Boolean;


implementation
//==============================================================================
uses Integers_Primes, HIT_Hash, HIT_Rand;

////////////////////////////////////////////////////////////////////////////////
// CONSTANTS
////////////////////////////////////////////////////////////////////////////////

const
  //----- Binary size of the "checking" fingerprint
  OTP_CHECK_SIZE = 160;
  //----- Hash functions G and H (Indexed RIPEMD-160 function)
  OTP_HFUNC_G = 1;
  OTP_HFUNC_H = OTP_HFUNC_G + 1; //Must be different from "G"

////////////////////////////////////////////////////////////////////////////////
// ROUTINES
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Return TRUE iff the P pointers are distinct
//------------------------------------------------------------------------------
function DistinctPointers(const P: array of Pointer): Boolean;
  var i, j : Longint;
begin
  Result := true;
  for i := 0 to High(P)-1 do
    for j := i+1 to High(P) do
      if P[i] = P[j] then
      begin
        Result := false;
        Exit;
      end;
end;

//------------------------------------------------------------------------------
// Generate the factors P and Q and the modulus N
// -> NSize, binary size of N
// -> rg, initialized secure random generator
// <- P, factor, prime
// <- Q, factor, prime, Q <> P
// <- N = P * Q
//------------------------------------------------------------------------------
procedure GetFactorsAndModulus(P,Q,N: TInteger; NSize: Longint; rg: TIntCSRG);
  var s : Longint;
begin
  // P and Q are strong pseudoprimes for the bases [2,3,5,7,11]
  // P and Q are equal to 2 mod 3 so that 3 can be the encryption exponent

  //----- Generate the prime P
  s := NSize shr 1;
  rg.Rnd(P,s);
  P.Bit[s-2] := true; //So that N.BitSize = NSize
  IMakePrime(P,[2,3,5,7,11],2,3);

  //----- Generate the prime Q
  s := NSize - P.BitSize;
  repeat
    rg.Rnd(Q,s);
    Q.Bit[s-2] := true; //So that N.BitSize = NSize
    IMakePrime(Q,[2,3,5,7,11],2,3);
  until (Q.BitSize = s) and not IEqu(P,Q); //Always true on first iterate

  //----- N := P * Q
  ISet(N,P);
  IMul(N,Q);
end;

//==============================================================================
// Compute the public and secret keys for the OTP-DRSA scheme
// -> NSize, binary size of the modulus N
// <- N, modulus (product of 2 primes)
// <- D, decryption exponent
//------------------------------------------------------------------------------
// In real-life application NSize should be at least equal to 1024
//------------------------------------------------------------------------------
// Public key: N,3
// Secret key: N,D
//==============================================================================
procedure OTPDRSAKeys(N, D: TInteger; NSize: Longint);
  var
    P, Q  : TInteger;
    rg    : TIntCSRG; //Secure Random Generator (See HIT_Rand.pas)
    OldCL : Boolean;
begin
  //----- Check
  if N = D then HITRaiseDuplicatePtr('OTPDRSAKeys','N,D');

  rg := nil;
  P := nil;
  Q := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- Create a secure random generator
    rg := TIntCSRG.Create;
    //----- Create integers
    P := TInteger.Create;
    Q := TInteger.Create;

    repeat
      //----- Factors and modulus
      GetFactorsAndModulus(P,Q,N,NSize,rg);

      //----- Carmichael Lambda-function L = LCM(P-1,Q-1)
      IDec(P);
      IDec(Q);
      ILCM(P,Q); //P := LCM(P-1,Q-1)

      //----- Exponent D (E = 3)
      ISet(D,3);
      IInvMod(D,P);
    until IGCD1(D,N); //If not, we are somewhat unlucky...

  finally
    //----- Free the generator
    rg.Free;
    //----- Free the integers
    P.Free;
    Q.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// Compute the public and secret keys for the OTP-DRSA scheme (using the
// Chinese Remainder Theorem)
// -> NSize, binary size of the modulus N
// <- N, modulus (product of 2 primes)
// <- (DP,DQ,UP,UQ,P,Q), decryption exponents and modulus factors
//------------------------------------------------------------------------------
// In real-life application NSize should be at least equal to 1024
//------------------------------------------------------------------------------
// Public key: N,3
// Secret key: N,DP,DQ,UP,UQ,P,Q
//==============================================================================
procedure OTPDRSAKeysCRT(N,DP,DQ,UP,UQ,P,Q: TInteger; NSize: Longint);
  var
    rg    : TIntCSRG; //Secure Random Generator (See HIT_Rand.pas)
    OldCL : Boolean;
begin
  //----- Check
  if not DistinctPointers([N,DP,DQ,UP,UQ,P,Q]) then
    HITRaiseDuplicatePtr('OTPDRSAKeysCRT','N,DP,DQ,UP,UQ,P,Q');

  rg := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- Create a secure random generator
    rg := TIntCSRG.Create;

    repeat
      //----- Factors and modulus
      GetFactorsAndModulus(P,Q,N,NSize,rg);

      //----- Carmichael Lambda-function L = LCM(P-1,Q-1)
      ISet(UP,P);
      ISet(UQ,Q);
      IDec(UP);
      IDec(UQ);
      ISet(DQ,UP);
      ILCM(DQ,UQ); //DQ := LCM(P-1,Q-1)

      //----- Exponent D (E = 3)
      ISet(DP,3);
      IInvMod(DP,DQ);
    until IGCD1(DP,N); //If not, we are somewhat unlucky...

    //----- Decryption exponents (here DP contains D)
    ISet(DQ,DP);
    IMod(DP,UP); //DP := D mod (P-1)
    IMod(DQ,UQ); //DQ := D mod (Q-1)
    //----- Chinese Remainder Theorem
    ISet(UQ,P);
    ISet(UP,Q);
    IInvMod(UQ,UP); //I = P**(-1) mod Q
    ISub(UP,UQ);    //-I mod Q (always positive)
    IMul(UP,P);
    IInc(UP);   //Up = (((0-1) * I) mod Q) * P + 1
    IMul(UQ,P); //Uq = (((1-0) * I) mod Q) * P + 0

  finally
    //----- Free the generator
    rg.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// OTP-DRSA: Encryption of a message
// -> M, message to encrypt (must be a nonnegative integer)
// -> (N,3), public key of the receiver
// <- (A,B,H), encrypted message
//------------------------------------------------------------------------------
// This routine works for both methods: Standard and CRT
//==============================================================================
procedure OTPDRSAEncrypt(A,B,H, M, N: TInteger);
  var
    rg    : TIntCSRG; //Secure Random Generator (See HIT_Rand.pas)
    s     : Longint;
    OldCL : Boolean;
begin
  //----- Check
  if not DistinctPointers([A,B,H,M,N]) then
    HITRaiseDuplicatePtr('OTPDRSAEncrypt','A,B,H,M,N');
  if M.Neg then HITRaiseInvalidArg('OTPDRSAEncrypt','M < 0');

  rg := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- Create a secure random generator
    rg := TIntCSRG.Create;

    //----- K, random, 0 < K < N
    s := N.BitSize;
    repeat
      rg.Rnd(A,s+s,false);
      IMod(A,N);
    until not A.Is0;
    ISet(B,A);
    ISet(H,A);

    //----- A := K^3 mod N
    IPowMod(A,3,N);

    //----- B := g((K+1)^3 mod N) xor M
    IInc(B);
    IPowMod(B,3,N);
    s := M.BitSize;
    IH160(B,s,OTP_HFUNC_G); //Create a MSize mask
    B.Bit[s] := true; //B := 1 & B
    IXor(B,M);

    //----- H := h(K & B & M)
    ICat(H,B,0);
    ICat(H,M,0);
    IH160(H,OTP_CHECK_SIZE,OTP_HFUNC_H);

  finally
    //----- Free the random generator
    rg.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// OTP-DRSA: Decryption of a message
// -> (A,B,H), encrypted message
// -> (N,3,D), public and private keys of the receiver
// <- M, decrypted message
// <- TRUE if and only if (A,B,H) is a valid encrypted message
//==============================================================================
function OTPDRSADecrypt(M, A,B,H, N,D: TInteger): Boolean;
  var
    X, Y  : TInteger;
    s     : Longint;
    OldCL : Boolean;
begin
  //----- Check
  if not DistinctPointers([M,A,B,H,N,D]) then
    HITRaiseDuplicatePtr('OTPDRSADecrypt','M,A,B,H,N,D');

  X := nil;
  Y := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- X = Y = K := A^D mod N
    X := A.Clone;
    IPowMod(X,D,N);
    Y := X.Clone;

    //----- M := g((K+1)^3 mod N) xor B
    ISet(M,X);
    IInc(M);
    IPowMod(M,3,N);
    //----- Create the mask
    s := B.BitSize-1;
    IH160(M,s,OTP_HFUNC_G);
    M.Bit[s] := true;
    IXor(M,B);

    //----- Y := h(K & B & M)
    ICat(Y,B,0);
    ICat(Y,M,0);
    IH160(Y,OTP_CHECK_SIZE,OTP_HFUNC_H);

    //----- Check
    Result := IEqu(H,Y);

  finally
    //----- Free the numbers
    X.Free;
    Y.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// OTP-DRSA: Decryption of a message (using the Chinese Remainder Theorem)
// -> (A,B,H), encrypted message
// -> (N,3,DP,DQ,UP,UQ,P,Q), public and private keys of the receiver
// <- M, decrypted message
// <- TRUE if and only if (A,B,H) is a valid encrypted message
//==============================================================================
function OTPDRSADecryptCRT(M, A,B,H, N,DP,DQ,UP,UQ,P,Q: TInteger): Boolean;
  var
    X, Y  : TInteger;
    s     : Longint;
    OldCL : Boolean;
begin
  //----- Check
  if not DistinctPointers([M,A,B,H,N,DP,DQ,UP,UQ,P,Q]) then
    HITRaiseDuplicatePtr('OTPDRSADecryptCRT','M,A,B,H,N,DP,DQ,UP,UQ,P,Q');

  X := nil;
  Y := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- Create X and Y equal to A
    X := A.Clone;
    Y := A.Clone;

    //===== Chinese remainder theorem
    //----- X := A^DP mod P
    IPowMod(X,DP,P);
    //----- Y := A^DQ mod Q
    IPowMod(Y,DQ,Q);
    //----- X := (X*UP + Y*UQ) mod N
    IMul(X,UP);
    IMul(Y,UQ);
    IAdd(X,Y);
    IMod(X,N);

    //----- Y := X = K = A^D mod N
    ISet(Y,X);

    //----- M := B xor g((K+1)^3 mod N)
    ISet(M,X);
    IInc(M);
    IPowMod(M,3,N);
    s := B.BitSize-1;
    IH160(M,s,OTP_HFUNC_G);
    M.Bit[s] := true;
    IXor(M,B);

    //----- Y := h(K & B & M)
    ICat(Y,B,0);
    ICat(Y,M,0);
    IH160(Y,OTP_CHECK_SIZE,OTP_HFUNC_H);

    //----- Check
    Result := IEqu(H,Y);

  finally
    //----- Free the numbers
    X.Free;
    Y.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TEST ROUTINES
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// OTP-DRSA: Test the unit routines
// -> NSize, modulus binary size
// <- te, encryption running time (in ms)
// <- td, decryption running time (in ms)
// <- TRUE if all is right, FALSE if an error occured
//==============================================================================
function OTPDRSATest(NSize: Longint; var te, td: Extended): Boolean;
  var
    N, D, M, A, B, H, Z : TInteger;
    t                   : Int64;
begin
  N := nil; D := nil; M := nil;
  A := nil; B := nil; H := nil; Z := nil;
  try
    N := TInteger.Create;
    D := TInteger.Create;
    M := TInteger.Create;
    A := TInteger.Create;
    B := TInteger.Create;
    H := TInteger.Create;
    Z := TInteger.Create;

    //----- Create the keys
    OTPDRSAKeys(N,D,NSize);

    //----- Encrypt a random message (with a random length < 2 * (N length))
    IRnd(M, IRnd32 mod Longword(2*N.BitSize));
    t := HITTimer;
    OTPDRSAEncrypt(A,B,H,M,N);
    te := HITTimer(t);

    //----- Check decryption
    t := HITTimer;
    Result := OTPDRSADecrypt(Z,A,B,H,N,D);
    td := HITTimer(t);

    //----- Identical message ? (useless but...)
    if Result then Result := IEqu(M,Z);

  finally
    N.Free;
    D.Free;
    M.Free;
    A.Free;
    B.Free;
    H.Free;
    Z.Free;
  end;
end;

//==============================================================================
// OTP-DRSA: Test the unit routines (using the Chinese Remainder Theorem)
// -> NSize, modulus binary size
// <- te, encryption running time (in ms)
// <- td, decryption running time (in ms)
// <- TRUE if all is right, FALSE if an error occured
//==============================================================================
function OTPDRSATestCRT(NSize: Longint; var te, td: Extended): Boolean;
  var
    N, DP, DQ, UP, UQ, P, Q, M, A, B, H, Z : TInteger;
    t                                      : Int64;
begin
  N := nil;
  DP := nil; DQ := nil; UP := nil; UQ := nil;
  P := nil; Q := nil;
  M := nil; A := nil; B := nil; H := nil; Z := nil;
  try
    N := TInteger.Create;
    DP := TInteger.Create;
    DQ := TInteger.Create;
    UP := TInteger.Create;
    UQ := TInteger.Create;
    P := TInteger.Create;
    Q := TInteger.Create;
    M := TInteger.Create;
    A := TInteger.Create;
    B := TInteger.Create;
    H := TInteger.Create;
    Z := TInteger.Create;

    //----- Create the keys
    OTPDRSAKeysCRT(N,DP,DQ,UP,UQ,P,Q,NSize);

    //----- Encrypt a random message (with a random length < 2 * (N length))
    IRnd(M, IRnd32 mod Longword(2*N.BitSize));
    t := HITTimer;
    OTPDRSAEncrypt(A,B,H,M,N);
    te := HITTimer(t);

    //----- Check decryption
    t := HITTimer;
    Result := OTPDRSADecryptCRT(Z,A,B,H,N,DP,DQ,UP,UQ,P,Q);
    td := HITTimer(t);

    //----- Identical message ? (useless but...)
    if Result then Result := IEqu(M,Z);

  finally
    N.Free;
    DP.Free;
    DQ.Free;
    UP.Free;
    UQ.Free;
    P.Free;
    Q.Free;
    M.Free;
    A.Free;
    B.Free;
    H.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

