////////////////////////////////////////////////////////////////////////////////
// Williams P+1 Factorization Method.
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit WilliamsPplus1;

interface
uses Integers;

procedure WilliamsPP1(N, D: TInteger);

implementation
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// The Williams P+1 method works when a composite number N has a factor P such
// that P+1 contains only "small" factors.
//
// Example of a factorization:
// ---------------------------
//
// N = 026 594 572 556 522 948 670 918 847 096 912 483 704 946 297 857 416 498
//     119 772 902 488 685 247 143 079
//
// Running time = 2.573 s
// D = 001 (failed)
//
// Running time = 2.577 s
// D = 001 (failed)
//
// Running time = 0.093 s
// D = 383 623 607 960 291 186 711 346 659 317 951 365 471 934 676 993
//
// Note that N = D*E with E+1 = 2^3 * 839 * 7043 * 1248090631 * 1174983937399
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Try to factor N using the Williams P+1 method
// -> N, composite number
// <- D, factor of N if succeeded or D = 1 if failed
//==============================================================================
procedure WilliamsPP1(N, D: TInteger);
  const
    ILOOP = 16;
    MAX   = ILOOP * 250; //Compute at most V[MAX!]
  var
    W, X     : TInteger;
    i, Count : Longword;
  //--------------------------------------------------------------------------
  // Calculate the V[k] element of a Lucas sequence modulo N
  // k > 0
  // N, odd
  // V <- V[k] modulo N
  //--------------------------------------------------------------------------
  procedure VkModN(V: TInteger; k: Longword; N: TInteger);
    var i : Longint;
  begin
    if k = 1 then Exit; //V has the right value
    ISet(W,V);
    ISet(X,V);
    //----- W = V - 2
    ISqr(W);
    ISub(W,2);
    IMod(W,N);
    //===== Loop
    for i := IBitSize32(k)-2 downto 1 do
    if IBit32(k,i) then
    begin
      //----- V = VW - X
      IMul(V,W);
      IMod(V,N);
      ISub(V,X);
      if V.Neg then IAdd(V,N);
      //----- W = W - 2
      ISqr(W);
      IMod(W,N);
      ISub(W,2);
      if W.Neg then IAdd(W,N);
    end
    else
    begin
      //----- W := VW - X
      IMul(W,V);
      IMod(W,N);
      ISub(W,X);
      if W.Neg then IAdd(W,N);
      //----- V = V - 2
      ISqr(V);
      IMod(V,N);
      ISub(V,2);
      if V.Neg then IAdd(V,N);
    end;
    //===== Trailing bit
    if Odd(k) then
    begin
      //----- V = VW - X
      IMul(V,W);
      IMod(V,N);
      ISub(V,X);
      if V.Neg then IAdd(V,N);
    end
    else
    begin
      //----- V = V - 2
      ISqr(V);
      IMod(V,N);
      ISub(V,2);
      if V.Neg then IAdd(V,N);
    end;
  end;
  //--------------------------------------------------------------------------
begin
  W := nil;
  X := nil;
  try
    W := TInteger.Create;
    X := TInteger.Create;
    //----- Random seed D such that 2 < D < N
    repeat
      IRnd(D,N.BitSize+1);
      IMod(D,N);
    until D.BitSize > 2;
    //------------------------------------------------------------------------
    // Let F be a prime factor of N such that F+1 has only small factors. We
    // should have Jacobi(D-4,F) = -1. But we can't check that because F is
    // unknown. Fortunately, we have roughly a 50-50 chance that a randomly
    // choosen D will work. For this reason, the Williams' method should be
    // tried at least 3 or 4 times.
    //------------------------------------------------------------------------
    ISub(D,2);
    Count := 0;
    //----- Main loop --------------------------------------------------------
    // Compute the V[k](D,1) element of a Lucas sequence for k = Count! and
    // check whether GCD(V[Count!]-2, N) <> 1.
    // A Lucas sequence V[k](P,Q) is specified by its (P,Q) parameters. Here,
    // we use the fact that, when Q = 1, V[a*b](P,1) = V[a](V[b](P,1),1).
    //------------------------------------------------------------------------
    while IGCD1(D,N) and (Count <= MAX) do
    begin
      IAdd(D,2);
      for i := 1 to ILOOP do
      begin
        Inc(Count);
        VkModN(D,Count,N);
      end;
      ISub(D,2);
    end;
    IGCD(D,N);
  finally
    W.Free;
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

