unit WinGImag;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Dialogs, StdCtrls, ExtCtrls, WinG, Dib256, DsgnIntf, Pal256;

type
   TFileNameProperty = class(TStringProperty)
      function GetAttributes: TPropertyAttributes; override;
      procedure Edit; override;
   end;

   ViewMode = (All, Portion);

   TView = class(TPersistent)
   private
      FMode: ViewMode;
      FLeft: Word;
      FTop: Word;
      FWidth: Word;
      FHeight: Word;
      FOnChange: TNotifyEvent;

      procedure SetMode( Value: ViewMode );
      procedure SetLeft( Value: Word );
      procedure SetTop( Value: Word );
      procedure SetWidth( Value: Word );
      procedure SetHeight( Value: Word );
   public
      procedure SetView( L, T, W, H: Word );
   published
      property Mode: ViewMode read FMode write SetMode;
      property Left: Word read FLeft write SetLeft;
      property Top: Word read FTop write SetTop;
      property Width: Word read FWidth write SetWidth;
      property Height: Word read FHeight write SetHeight;

      property OnChange: TNotifyEvent read FOnChange write FOnChange;
   end;

    TWinGImage = class(TGraphicControl)
    private
      FAutoSize: Boolean;
      FPalette: TPalette256;
      FFileName: TFileName;
      FView: TView;
      FWinG: TWinG;
      FOnPaint: TNotifyEvent;

      procedure SetAutoSize( Value: Boolean );
      procedure SetFileName( Value: TFileName );

      procedure ViewChange( Sender: TObject );
      procedure WinGChange( Sender: TObject );
      procedure NewPalette( Sender: TObject );
      procedure IdentityChange( Sender: TObject );
      procedure BeforeAnimatePalette( Sender: TObject );
      procedure AfterAnimatePalette( Sender: TObject; StartIndex: Integer;
         ColorCount: Integer);

      procedure SetWinGPalette( StartIndex: Integer; ColorCount: Integer );
      procedure DeleteImage;

    protected
      function GetPalette: HPALETTE; override;
      procedure Paint; override;
      procedure Draw;

    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;

      procedure LoadFromFile;
      procedure SaveToFile( OutputFileName: TFileName );

      property Canvas;

    published
      property Cursor;
      property Left;
      property Top;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDrag;

      property AutoSize: Boolean read FAutoSize write SetAutoSize;
      property WinG: TWinG read FWinG;
      property FileName: TFileName read FFileName write SetFileName;
      property Palette: TPalette256 read FPalette write FPalette;
      property View: TView read FView write FView;
      property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
   end;

   Procedure ClearSystemPalette;
   Procedure Register;


implementation


procedure Register;
begin
   RegisterPropertyEditor( TypeInfo(TFileName), TWinGImage,'',
      TFileNameProperty );
	RegisterComponents('Additional', [TWinGImage]);
end;



{**** methods ****}
constructor TWinGImage.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);

{REG}

  WinGCreateDC();

  Width := 32;
  Height := 32;

   FView := TView.Create;
   FWinG := TWinG.Create(1,1);
   FPalette := TPalette256.Create;
   FAutoSize := True;
   FileName := ''; // don't replace by FFileName, inits file name

   { component is considered opaque, so be carefull to draw all
    the area }
   ControlStyle := ControlStyle + [csOpaque];

   FView.OnChange := ViewChange;
   FWinG.OnChange := WinGChange;
   FPalette.OnNewPalette := NewPalette;
   FPalette.OnIdentityChange := IdentityChange;
   FPalette.OnBeforeAnimatePalette := BeforeAnimatePalette;
   FPalette.OnAfterAnimatePalette := AfterAnimatePalette;
end;


destructor TWinGImage.Destroy;
begin
  FWinG.Free;
  FPalette.Free;
  FView.Free;

  inherited Destroy;
end;


procedure TWinGImage.LoadFromFile;
var
   dib: TDIB256;
   W, H: Word;
begin
   // checks the filename if it's empty
   if (FFileName = '') or (FFileName = '(none)') then
      exit;

   // create DIB to do the file reading
   dib := TDIB256.Create;
   try
      // read file
      dib.LoadFromFile( FFileName );

      // init. WinG DC size with DIB's
      if (FAutoSize = True) then
      begin
         FWinG.Width := dib.Width;
         FWinG.Height := dib.Height;
      end;

      // calc. smaller dimensions between the two DIBs
      if (FWinG.Width < dib.Width) then
         W := FWinG.Width
      else
         W := dib.Width;
      if (FWinG.Height < dib.Height) then
         H := FWinG.Height
      else
         H := dib.Height;

      CopyDIBBits( FWinG.Pixels, dib.Pixels,
         W, H, FWinG.WidthBytes, dib.WidthBytes );

      FPalette.ReadFromRGBQuad( dib.ColorTable, 256 );
      FPalette.CreateHandle;

      // refresh image and check View values
      FView.Left := FView.Left;
      Invalidate;
   finally
      dib.Free;
   end;
end;


procedure TWinGImage.SaveToFile( OutputFileName: TFileName );
var
  dib: TDib256;
begin
   // create DIB to do the file reading
   dib := TDIB256.Create;
   try
      dib.Width := WinG.Width;
      dib.Height := WinG.Height;

      CopyDIBBits( dib.Pixels, WinG.Pixels,
        dib.Width, dib.Height, dib.WidthBytes, WinG.WidthBytes );

      Palette.WriteToRGBQuad( dib.ColorTable, 256 );

      // read file
      dib.SaveToFile( OutputFileName );
   finally
      dib.Free;
   end;
end;


function TWinGImage.GetPalette: HPALETTE;
begin
   Result := FPalette.Handle;
end;


procedure TWinGImage.Paint;
var
   hd: hDC;
   oldhPal: HPALETTE;
begin
   hd := Canvas.Handle;
   if (hd <> 0) then
      if (FPalette.Handle <> 0)
      and ((FFileName <> '(none)') or not(csDesigning in ComponentState)) then
      begin
         oldhPal := SelectPalette( hd, FPalette.Handle, false );
         RealizePalette( hd );

         Draw;

         if Assigned(FOnPaint) then
           FOnPaint( self );

         SelectPalette( hd, oldhPal, True );
      end
      else if csDesigning in ComponentState then
         with Canvas do
         begin
           Pen.Style := psDash;
           Brush.Style := bsClear;
           Rectangle(0, 0, Width, Height);
         end;
end;


procedure TWinGImage.Draw;
begin
   if (View.Mode = All) then
      WinGStretchBlt( Canvas.Handle, 0, 0, Width, Height, FWinG.Handle,
       0, 0, FWinG.Width, FWinG.Height )
   else
      WinGStretchBlt( Canvas.Handle, 0, 0, Width, Height, FWinG.Handle,
       View.Left, View.Top, View.Width, View.Height );
end;


{**** properties ****}
procedure TWinGImage.SetAutoSize( Value: Boolean );
begin
   FAutoSize := Value;

   if not(csReading in ComponentState) and (FAutoSize = True) then
      LoadFromFile;
end;



procedure TWinGImage.SetFileName( Value: TFileName );
begin
   DeleteImage;

   FFileName := '(none)';
   if (Value <> '') then
   begin
      FFileName := Value;
      LoadFromFile;
   end;
   Invalidate;
end;


{**** private methods ****}
procedure TWinGImage.ViewChange( Sender: TObject );
begin
   if not(csReading in ComponentState) then
   begin
      if (FView.Left >= FWinG.Width) then
         FView.Left := FWinG.Width-1;
      if (FView.Top >= FWinG.Height) then
         FView.Top := FWinG.Height-1;
      if (FView.Width+FView.Left > FWinG.Width) then
         FView.Width := FWinG.Width - FView.Left;
      if (FView.Height+FView.Top > FWinG.Height) then
         FView.Height := FWinG.Height - FView.Top;

      Invalidate;
   end;
end;


procedure TWinGImage.WinGChange( Sender: TObject );
begin
   if not(csReading in ComponentState) then
   begin
      LoadFromFile;

      Invalidate;
   end;
end;


procedure TWinGImage.NewPalette( Sender: TObject );
begin
   SetWinGPalette( 0, 256 );
   Invalidate;
end;


procedure TWinGImage.IdentityChange( Sender: TObject );
begin
   LoadFromFile;
end;


procedure TWinGImage.BeforeAnimatePalette( Sender: TObject );
begin
   SelectPalette( Canvas.Handle, FPalette.Handle, False );
end;

procedure TWinGImage.AfterAnimatePalette( Sender: TObject; StartIndex: Integer;
   ColorCount: Integer );
begin
   SetWinGPalette( StartIndex, ColorCount );
   Draw;
end;


procedure TWinGImage.SetWinGPalette( StartIndex: Integer; ColorCount: Integer );
var
   Cnt: Integer;
   p: PRGBQUAD256;
begin
   p := PRGBQUAD256(FWinG.ColorTable);
   For Cnt := StartIndex to StartIndex+ColorCount-1 do
   begin
      p[Cnt].rgbBlue := FPalette.Entries^[Cnt].peBlue;
      p[Cnt].rgbRed := FPalette.Entries^[Cnt].peRed;
      p[Cnt].rgbGreen := FPalette.Entries^[Cnt].peGreen;
      p[Cnt].rgbReserved := 0;
   end;

   // Init. WinG DC palette
   WinGSetDIBColorTable( FWinG.Handle, StartIndex, ColorCount,
      Pointer(@p[StartIndex]) );
end;


procedure TWinGImage.DeleteImage;
var
   p: Pointer;
   nb: Longint;
begin
   p := FWinG.Pixels;
   nb := Longint(FWinG.Height) * Longint(FWinG.Width);
   asm
      push edi
      mov edi, p
      mov ecx, nb
      mov al, 0
      rep stosb
      pop edi
   end;
end;

{****** Property Editor *****}
function TFileNameProperty.GetAttributes: TPropertyAttributes;
begin
   Result := inherited GetAttributes + [paDialog] - [paSubProperties];
end;


procedure TFileNameProperty.Edit;
var
   Dialog: TOpenDialog;
Begin
   // changed Application for nil so that Forms isn't needed, and
   // works well in Japanese version
   Dialog := TOpenDialog.Create( nil );
   try
      Dialog.Filter := 'Bitmap files|*.BMP|DIB Files|*.DIB';
      if (Dialog.Execute = True) then
         SetStrValue( Dialog.FileName );
   finally
      Dialog.Free;
   end;
End;




procedure TView.SetView( L, T, W, H: Word );
begin
   FLeft := L;
   FTop := T;
   FWidth := W;
   FHeight := H;
   if Assigned(FOnChange) then OnChange(Self);
end;

procedure TView.SetMode( Value: ViewMode );
begin
   FMode := Value;
   if Assigned(FOnChange) then OnChange(Self);
end;

procedure TView.SetLeft( Value: Word );
begin
   FLeft := Value;
   if Assigned(FOnChange) then OnChange(Self);
end;

procedure TView.SetWidth( Value: Word );
begin
   FWidth := Value;
   if Assigned(FOnChange) then OnChange(Self);
end;

procedure TView.SetTop( Value: Word );
begin
   FTop := Value;
   if Assigned(FOnChange) then OnChange(Self);
end;

procedure TView.SetHeight( Value: Word );
begin
   FHeight := Value;
   if Assigned(FOnChange) then OnChange(Self);
end;




Procedure ClearSystemPalette;
{/*** A dummy palette setup }
var
	Counter: Integer;
   Palette: record
      Version: Word;
      NumberOfEntries: Word;
      aEntries: array[0..255] of TPaletteEntry;
   end;
   ScreenPalette: hPalette;
	ScreenDC: hDC;

begin
	Palette.Version := $300;
   Palette.NumberOfEntries := 256;

	{ *** Reset everything in the system palette to black }
  	for Counter := 0 to 255 do
   begin
      Palette.aEntries[Counter].peRed := 0;
		Palette.aEntries[Counter].peGreen := 0;
    	Palette.aEntries[Counter].peBlue := 0;
    	Palette.aEntries[Counter].peFlags := PC_NOCOLLAPSE;
   end;

	{ *** Create, select, realize, deselect, and delete the palette }
	ScreenDC := GetDC( 0 );
	ScreenPalette := CreatePalette( pLogPalette(@Palette)^ );

  	if ScreenPalette <> 0 then
  	begin
    	ScreenPalette := SelectPalette(ScreenDC, ScreenPalette, FALSE);
    	RealizePalette( ScreenDC );
    	SelectPalette(ScreenDC, ScreenPalette, FALSE);
    	DeleteObject( ScreenPalette );
  	end;

	ReleaseDC(0, ScreenDC);
end;


end.
