{*******************************************************************************
***  Unit : Component TBuildInfo
***  Author : Morten Hyseth
***  Construction date : 24.06.99
***  Description : Retrieves build information into component properties
***  Edit log : v1.0
***
***  Usage : call the "GetFileInfo(Filename : string)" function to load
***          information into properties. Use full path and extention, when
***          sending in the filename parameter.
*******************************************************************************}

unit infBuildInfo;

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, dsgnintf;

Const
  NoString = '';
  cSlash   = '\';
  cPre     = 'Prerelease';
  cDebug   = 'Debug';
  cPatched = 'Patch';
  cPrivate = 'Private';
  cSpecial = 'Special';

type
  { property editors and stuff :}
  TTestPropEd = class(TClassProperty)
  public
    function GetAttributes : TPropertyAttributes; override;
  end;

  TBuildInfo = class(TPersistent)
  private
    FisPrerelease : boolean;
    FisDebug      : boolean;
    FisPatch      : boolean;
    FisPrivate    : boolean;
    FisSpecial    : boolean;
  published
    property isPrerelease  : boolean read FisPrerelease write FisPrerelease;
    property isDebug       : boolean read FisDebug write FisDebug;
    property isPatch       : boolean read FisPatch write FisPatch;
    property isPrivate     : boolean read FisPrivate write FisPrivate;
    property isSpecial     : boolean read FisSpecial write FisSpecial;
  end;

  { usual class defenitions :}

  TInfGetVersion = class(TComponent)
  private
    { Private declarations }
    fBuildInfo : TBuildInfo;
    fFileDate,
    fSignatur,
    fStructVer,
    fFileType,
    fOpSystem  : String;
    FFBuffer   : PVSFixedFileInfo;
    fCompanyName,
    fFileDescription,
    fFileVersion,
    fInternalName,
    fLegalCopyright,
    fLegalTrademark,
    fOriginalFilename,
    fProductName,
    fProductVersion,
    fComments,
    FOSVersion : string;
    function GetAllFileVersionInfo(Filename : String) : boolean;
    function GetRootInfo(Buffer : String) : string;
    procedure SetOSVersion(const Value: string);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    function GetFileInfo(Filename : string) : boolean;
  published
    { Published declarations }
    property CompanyName      : string read fCompanyName write fCompanyName;
    property FileDescription  : string read fFileDescription write fFileDescription;
    property FileVersion      : string read fFileVersion write fFileVersion;
    property InternalName     : string read fInternalName write fInternalName;
    property LegalCopyright   : string read fLegalCopyright write fLegalCopyright;
    property LegalTrademark   : string read fLegalTrademark write fLegalTrademark;
    property OriginalFilename : string read fOriginalFilename write fOriginalFilename;
    property ProductName      : string read fProductName write fProductName;
    property ProductVersion   : string read fProductVersion write fProductVersion;
    property Comments         : string read fComments write fComments;
    property OSVersion        : string read FOSVersion write SetOSVersion;
    property BuildInfo        : TBuildInfo read fBuildInfo write fBuildInfo;
    property Signatur         : String read fSignatur write fSignatur;
    property StructVer        : String read fStructVer write fStructVer;
    property FileType         : String read fFileType write fFileType;
    property OpSystem         : String read fOpSystem write fOpSystem;
  end;

procedure Register;

implementation

{*******************************************************************************
***  Implementation of the
***  TinfGetVersion component
*******************************************************************************}

const
  FormatStr            = '%s%.4x%.4x\%s%s';
  cE_obtain : PChar    = 'Could not obtain version information.';
  cE_error  : PChar    = 'Error';
  cTransString : PChar = '\VarFileInfo\Translation';
  cStringInfo  : PChar = '\StringFileInfo\';
  {information constants :}
  cCompName            = 'CompanyName';
  cFileDesc            = 'FileDescription';
  cFileVer             = 'FileVersion';
  cIntName             = 'InternalName';
  cLegCopy             = 'LegalCopyright';
  cLegTrad             = 'LegalTrademarks';
  cOrgFile             = 'OriginalFilename';
  cProdName            = 'Productname';
  cProdVer             = 'ProductVersion';
  cComments            = 'Comments';

function TInfGetVersion.GetFileInfo(Filename : string) : boolean;
{ published procedure, checkes if filename is empty and
  if the file exists }
begin
  IF ((Filename <> NoString) and FileExists(Filename)) then
    Result := GetAllFileVersionInfo(Filename)
  else
    Result := False;
end;

procedure TInfGetVersion.SetOSVersion(const Value: string);
begin
  FOSVersion := Value;
end;

function TInfGetVersion.GetRootInfo(Buffer : String) : string;

  function GetOpSys(Value : PVSFixedFileInfo) : string;
  begin
    With Value^ do
    begin
      case dwFileOs of
        VOS_DOS   : Result := 'Designed for MS-DOS';
        VOS_OS216 : Result := 'Designed for 16-bit OS/2';
        VOS_OS232 : Result := 'Designed for 32-bit OS/2';
        VOS_NT    : Result := 'Designed for Windows NT';
//      VOS_WINDOWS16 : Result := 'Designed for 16-bit Windows';
//      VOS_WINDOWS32 : Result := 'Designed for 32-bit Windows';
//      VOS_PM16      : Result := 'Designed for 16-bit PM';
//      VOS_PM32      : Result := 'Designed for 32-bit PM';
        VOS_DOS_WINDOWS16 : Result := 'Designed for 16-bit Windows, running on DOS';
        VOS_DOS_WINDOWS32 : Result := 'Designed for 32-bit Windows, running on DOS';
        VOS_OS216_PM16    : Result := 'Designed for 16-bit PM, running on 16-bit OS/2';
        VOS_OS232_PM32    : Result := 'Designed for 32-bit PM, running on 32-bit OS/2';
        VOS_NT_WINDOWS32  : Result := 'Designed for Win32 API, running on Windows/NT';
      else
        Result := 'Unknown';
      end;
    end;
  end;

var
  Value   : PVSFixedFileInfo;
  CurSize : Cardinal;
begin
  Result := NoString;
  IF VerQueryValue(PChar(Buffer), cSlash, pointer(Value), CurSize) then
  begin
    with Value^ do
    begin
      IF (dwFileFlagsMask and dwFileFlags and VS_FF_PRERELEASE) <> 0 then
        Buildinfo.isPrerelease := True else Buildinfo.isPrerelease := False;
      IF (dwFileFlagsMask and dwFileFlags and VS_FF_DEBUG) <> 0 then
        Buildinfo.isDebug := True else Buildinfo.isDebug := False;
      IF (dwFileFlagsMask and dwFileFlags and VS_FF_PATCHED) <> 0 then
        Buildinfo.isPatch := True else Buildinfo.isPatch := False;
      IF (dwFileFlagsMask and dwFileFlags and VS_FF_PRIVATEBUILD) <> 0 then
        Buildinfo.isPrivate := True else Buildinfo.isPrivate := False;
      IF (dwFileFlagsMask and dwFileFlags and VS_FF_SPECIALBUILD) <> 0 then
        Buildinfo.isSpecial := True else Buildinfo.isSpecial := False;
      Signatur  := IntToStr(Value^.dwSignature);
      StructVer := IntToStr(Value^.dwStrucVersion);
      IF Value^.dwFileType = VFT_APP then FileType             := 'Application'
      else if Value^.dwFileType = VFT_DLL then FileType        := 'DLL'
      else if Value^.dwFileType = VFT_DRV then FileType        := 'Device driver'
      else if Value^.dwFileType = VFT_FONT then FileType       := 'Font file'
      else if Value^.dwFileType = VFT_VXD then FileType        := 'Virtual device'
      else if Value^.dwFileType = VFT_STATIC_LIB then FileType := 'Static library'
      else FileType := 'Unknown';
      OpSystem := GetOpSys(Value);
     end; //with
  end;
end;

function TInfGetVersion.GetAllFileVersionInfo(Filename : String) : boolean;
{ Gets fileinformation
  first calls InitVersion, then GetVersion for each thing to
  retrive. }
var
  Buffer    : String;
  fInfoSize : DWORD;

  function InitVersion : boolean;
  var
    FilenamePointer : PChar;
  begin
    Result          := True;
    FilenamePointer := PChar(Filename);
    fInfoSize       := GetFileVersionInfoSize(FilenamePointer, fInfoSize);
    IF fInfoSize > 0 then
    begin
      SetLength(Buffer, fInfoSize);
      IF not GetFileVersionInfo(FilenamePointer, 0, fInfoSize, PChar(Buffer)) then
      begin
        Result := False;
        Application.MessageBox(cE_obtain, cE_error, MB_OK);
      end;
    end; //if
  end; //InitVersion

  function GetVersion(whatToGet : String) : string;
  var
    tmpVersion : String;
    Len, Len2  : DWORD;
    Value      : PChar;
    temp       : PLongInt;
    tempStr    : String;
  begin
    Result := NoString;
    IF fInfoSize > 0 then
    begin
      SetLength(tmpVersion, 200);
      Value := @tmpVersion;
      {retrieves translation information (used to retrive strings) :}
      VerQueryValue(PChar(Buffer), cTransString, Pointer(temp), Len2);
      tempStr := Format(FormatStr, [cStringInfo, LoWord(temp^), HiWord(temp^), whattoget, #0]);
      {string retrieval :}
      IF VerQueryValue(PChar(Buffer), PChar(tempStr), Pointer(Value), Len) then
        Result := Value;
    end; // if
  end; //getversion

begin
  {initialization :}
  Buffer := NoString;
  Result := True;
  try
    Result := InitVersion;
    GetRootInfo(Buffer);
    {retrieve information :}
    IF Result then
    begin
      fCompanyName      := GetVersion(cCompName);
      fFileDescription  := GetVersion(cFileDesc);
      fFileVersion      := GetVersion(cFileVer);
      fInternalName     := GetVersion(cIntName);
      fLegalCopyright   := GetVersion(cLegCopy);
      fLegalTrademark   := GetVersion(cLegTrad);
      fOriginalFilename := GetVersion(cOrgFile);
      fProductName      := GetVersion(cProdName);
      fProductVersion   := GetVersion(cProdVer);
      fComments         := GetVersion(cComments);
    end;
  except
    Result := False;
  end;
end; //infGetVersion

constructor TInfGetVersion.Create(AOwner: TComponent);
begin
  inherited;
  fBuildInfo := TBuildInfo.Create;
end;

destructor TInfGetVersion.Destroy;
begin
  fBuildInfo.Free;
  inherited;
end;

{*******************************************************************************
*** Here starts component registrating
*******************************************************************************}

function TTestPropEd.GetAttributes : TPropertyAttributes;
begin
  Result := [paSubProperties];
end;


procedure Register;
begin
  RegisterComponents('Informatics', [TInfGetVersion]);
  RegisterPropertyEditor(TypeInfo(TBuildInfo), nil, '', TTestPropEd);
end;

end.
