{*******************************************************************************
***  Unit : Component infMutex
***  Author : Morten Hyseth
***  Construction date : 14.04.99
***  Description : Checks if an instance of the program is allready running
***  Edit log : v1.0
***
***  Usage :
***  1. Run the function AlreadyRunning to check if your program is running
***     in another instance. It actually checks if the mutex name (property
***     MutexName) already exists. Put it in an startup event.
***  2. To free/unregister the mutex(name) call the remove mutex.
***     Remember if you change the MutexName-property after you have
***     run the AlreadyRunning function, it won't work (it doesen't do anything)
***     Simple rule : don't change the MutexName property runtime if you
***     don't know what you're doing (and my guess is you don't).
***  3. Designtime you're free to change the Mutexname to something more
***     relevant than the default.
***  4. You should call the RemoveMutex function before closing you application
***  5. The funtions have a lot of other usage possibilities than
***     checking if an application is runing or not.
***
***  Ex. check if your app is already running :
***
***    if InfMutex1.AlreadyRunning(true) then
***    begin
***       showmessage('I am running already');
***       Application.Terminate;
***    end;
***
*******************************************************************************}

unit infMutex;

interface

uses
  Windows, Classes, SysUtils;

Const
  NoString       = '';
type
  { usual class defenitions :}
  { infMutex def.
    Current exported functions :
      AlreadyRunning - Checks if the mutex exists, and creates one if it
                       doesn't (if set in parameter).
      RemoveMutex    - Removes the mutex.
  }
  TInfMutex = class(TComponent)
  private
    { Private declarations }
    FMutexName: String;
    procedure SetMutexName(const Value: String);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner:TComponent);override;
  published
    { Published declarations }
    property MutexName : String read FMutexName write SetMutexName;
    function AlreadyRunning(DoCreateMutex : Boolean) : Boolean;
    function RemoveMutex : Boolean;
  end;

procedure Register;

implementation

uses
  Winsock, registry, shlObj, ActiveX, ComObj;

{*******************************************************************************
***  Implementation of the
***  infMutex component
*******************************************************************************}

var
  Muthandle : THandle;

function TInfMutex.AlreadyRunning(DoCreateMutex : Boolean) : Boolean;
{*******************************************************************************
*** Returns true if another instance of the application is already running.
*** Returns false if not.  (Makes sense, huh ?)
*** If createmutex is true, it creates a mutex if it doesen't exist
*******************************************************************************}
var
  localMutHandle : THandle;
begin
  IF MutexName <> NoString then
  begin
    localMuthandle := OpenMutex(MUTEX_ALL_ACCESS, False, PChar(MutexName));
    try
      IF localMuthandle = 0 then
      begin
        IF (DoCreateMutex = True) then
          MutHandle := CreateMutex(nil, false, PChar(MutexName));
        Result    := False;
      end
      else
        Result    := True;
    finally
      CloseHandle(localMutHandle);
    end;//try..finally
  end; //if mutexname <> nostring
end;

constructor TInfMutex.Create(AOwner: TComponent);
begin
  inherited;
  MutexName := FormatDateTime('"MHO"ddmmhhmm', now);
end;

function TInfMutex.RemoveMutex : Boolean;
{******************************************************************************
*** Removes the mutex if it exists.
*** You should call this function before closing you application
*******************************************************************************}
begin
  try
    IF Muthandle <> 0 then
    begin
      CloseHandle(MutHandle);
      MutHandle := 0;
    end;
    Result := True;
  except
    Result := False;
  end;
end;

procedure TInfMutex.SetMutexName(const Value: String);
begin
  FMutexName := Value;
end;

{*******************************************************************************
*** Here starts component registrating
*******************************************************************************}

procedure Register;
begin
  RegisterComponents('Informatics', [TInfMutex]);
end;


end.
